@@ -17,6 +17,7 @@ function display_help()
   echo "    [-r]--relocatable] create a package to support relocatable ROCm"
   echo "    [-c|--clients] build library clients too (combines with -i & -d)"
   echo "    [-g|--debug] -DCMAKE_BUILD_TYPE=Debug (default is =Release)"
+  echo "    [-k|--relwithdebinfo] -DCMAKE_BUILD_TYPE=RelWithDebInfo"
   echo "    [--build-dir] build directory (default is ./build)"
   echo "    [--host] build library for host backend only"
   echo "    [--no-openmp] build library without OpenMP"
@@ -24,9 +25,8 @@ function display_help()
   echo "    [--static] build static library"
   echo "    [--compiler] compiler to build with"
   echo "    [--verbose] print additional cmake build information"
-  echo "    [--address-sanitizer] Build with address sanitizer enabled. Uses hipcc as compiler"
+  echo "    [--address-sanitizer] Build with address sanitizer enabled. Uses amdclang++ as compiler"
   echo "    [--codecoverage] build with code coverage profiling enabled"
-  echo "    [--rm-legacy-include-dir] Remove legacy include dir Packaging added for file/folder reorg backward compatibility"
 }
 
 # This function is helpful for dockerfiles that do not have sudo installed, but the default user is root
@@ -49,25 +49,13 @@ supported_distro( )
   esac
 }
 
-# This function is helpful for dockerfiles that do not have sudo installed, but the default user is root
-check_exit_code( )
-{
-  if (( $1 != 0 )); then
-    exit $1
-  fi
-}
-
 # This function is helpful for dockerfiles that do not have sudo installed, but the default user is root
 elevate_if_not_root( )
 {
-  local uid=$(id -u)
-
-  if (( ${uid} )); then
-    sudo $@
-    check_exit_code
+  if (( $EUID )); then
+    sudo "$@" || exit
   else
-    $@
-    check_exit_code
+    "$@" || exit
   fi
 }
 
@@ -138,13 +126,26 @@ install_packages( )
   local library_dependencies_ubuntu=( "make" "cmake" "pkg-config" )
   local library_dependencies_centos=( "epel-release" "make" "cmake3" "gcc-c++" "rpm-build" )
   local library_dependencies_fedora=( "make" "cmake" "gcc-c++" "libcxx-devel" "rpm-build" )
-  local library_dependencies_sles=( "make" "cmake" "gcc-c++" "libcxxtools9" "rpm-build" )
+  local library_dependencies_sles=( "make" "cmake" "gcc-c++" "rpm-build" )
 
   local client_dependencies_ubuntu=( "libboost-program-options-dev" )
   local client_dependencies_centos=( "boost-devel" )
   local client_dependencies_fedora=( "boost-devel" )
   local client_dependencies_sles=( "libboost_program_options1_66_0-devel" "pkg-config" "dpkg" )
 
+  if [[ ( "${ID}" == "sles" ) ]]; then
+    if [[ -f /etc/os-release ]]; then
+      . /etc/os-release
+
+      function version { echo "$@" | awk -F. '{ printf("%d%03d%03d%03d\n", $1,$2,$3,$4); }'; }
+      if [[ $(version $VERSION_ID) -ge $(version 15.4) ]]; then
+          library_dependencies_sles+=( "libcxxtools10" )
+      else
+          library_dependencies_sles+=( "libcxxtools9" )
+      fi
+    fi
+  fi
+
   # Host build
   if [[ "${build_host}" == false ]]; then
     library_dependencies_ubuntu+=( "libnuma1" )
@@ -267,6 +268,7 @@ build_mpi=false
 mpi_dir=deps/openmpi
 build_omp=true
 build_release=true
+build_release_debug=false
 build_dir=./build
 install_prefix=rocalution-install
 rocm_path=/opt/rocm
@@ -274,7 +276,6 @@ build_relocatable=false
 build_static=false
 build_address_sanitizer=false
 build_codecoverage=false
-build_freorg_bkwdcomp=true
 compiler=c++
 verb=false
 
@@ -285,7 +286,7 @@ verb=false
 # check if we have a modern version of getopt that can handle whitespace and long parameters
 getopt -T
 if [[ $? -eq 4 ]]; then
-  GETOPT_PARSE=$(getopt --name "${0}" --longoptions help,install,clients,dependencies,debug,build-dir:,host,no-openmp,mpi:,relocatable,codecoverage,static,compiler:,verbose,address-sanitizer,rm-legacy-include-dir --options hicgdr -- "$@")
+  GETOPT_PARSE=$(getopt --name "${0}" --longoptions help,install,clients,dependencies,debug,build-dir:,host,no-openmp,mpi:,relocatable,codecoverage,relwithdebinfo,static,compiler:,verbose,address-sanitizer,rm-legacy-include-dir --options hicgdrk -- "$@")
 else
   echo "Need a new version of getopt"
   exit 1
@@ -330,14 +331,15 @@ while true; do
         shift ;;
     --address-sanitizer)
         build_address_sanitizer=true
-        compiler=hipcc
+        compiler=amdclang++
+        shift ;;
+    -k|--relwithdebinfo)
+        build_release=false
+        build_release_debug=true
         shift ;;
     --codecoverage)
         build_codecoverage=true
         shift ;;
-    --rm-legacy-include-dir)
-        build_freorg_bkwdcomp=false
-        shift ;;
     --mpi)
         mpi_dir=${2}
         if [[ "${mpi_dir}" == off || "${mpi_dir}" == false || "${mpi_dir}" == 0 || "${mpi_dir}" == disabled ]]; then
@@ -347,8 +349,8 @@ while true; do
         fi
         shift 2 ;;
     --static)
-        #Forcing hipcc for static builds temporarily
-        compiler=${rocm_path}/bin/hipcc
+        #Forcing amdclang++ for static builds temporarily
+        compiler=${rocm_path}/bin/amdclang++
         build_static=true
         shift ;;
     --compiler)
@@ -375,6 +377,8 @@ printf "\033[32mCreating project build directory in: \033[33m${build_dir}\033[0m
 # ensure a clean build environment
 if [[ "${build_release}" == true ]]; then
   rm -rf ${build_dir}/release
+elif [[ "${build_release_debug}" == true ]]; then
+  rm -rf ${build_dir}/release-debug
 else
   rm -rf ${build_dir}/debug
 fi
@@ -455,6 +459,9 @@ pushd .
   if [[ "${build_release}" == true ]]; then
     mkdir -p ${build_dir}/release/clients && cd ${build_dir}/release
     cmake_common_options="${cmake_common_options} -DCMAKE_BUILD_TYPE=Release"
+  elif [[ "${build_release_debug}" == true ]]; then
+    mkdir -p ${build_dir}/release-debug/clients && cd ${build_dir}/release-debug
+    cmake_common_options+=("-DCMAKE_BUILD_TYPE=RelWithDebInfo")
   else
     mkdir -p ${build_dir}/debug/clients && cd ${build_dir}/debug
     cmake_common_options="${cmake_common_options} -DCMAKE_BUILD_TYPE=Debug"
@@ -498,16 +505,9 @@ pushd .
           echo "Code coverage is disabled in Release mode, to enable code coverage select either Debug mode (-g | --debug) or RelWithDebInfo mode (-k | --relwithdebinfo); aborting";
           exit 1
       fi
-      cmake_common_options="${cmake_common_options} -DBUILD_CODE_COVERAGE=ON"
+      cmake_common_options="${cmake_common_options} -DBUILD_CODE_COVERAGE=ON -DBUILD_SUPPORT_COMPLEX=OFF"
   fi
 
-  #Enable backward compatibility wrappers 
-  if [[ "${build_freorg_bkwdcomp}" == true ]]; then
-    cmake_common_options="${cmake_common_options} -DBUILD_FILE_REORG_BACKWARD_COMPATIBILITY=ON"
-  else
-    cmake_common_options="${cmake_common_options} -DBUILD_FILE_REORG_BACKWARD_COMPATIBILITY=OFF"
-  fi
- 
   # Verbose cmake
   if [[ "${verb}" == true ]]; then
     cmake_common_options="${cmake_common_options} -DBUILD_VERBOSE=ON"
@@ -529,23 +529,19 @@ pushd .
       -DCMAKE_MODULE_PATH="${rocm_path}/lib/cmake/hip ${rocm_path}/hip/cmake" \
       -DCMAKE_EXE_LINKER_FLAGS=" -Wl,--enable-new-dtags -Wl,--rpath,${rocm_path}/lib:${rocm_path}/lib64" \
       -DROCM_DISABLE_LDCONFIG=ON \
-      -DROCM_PATH="${rocm_path}" ../..
+      -DROCM_PATH="${rocm_path}" ../.. || exit
   else
-    CXX=${compiler} ${cmake_executable} ${cmake_common_options} ${cmake_client_options} -DCMAKE_INSTALL_PREFIX=${install_prefix} -DROCM_PATH=${rocm_path} ../..
+    CXX=${compiler} ${cmake_executable} ${cmake_common_options} ${cmake_client_options} -DCMAKE_INSTALL_PREFIX=${install_prefix} -DROCM_PATH=${rocm_path} ../.. || exit
   fi
 
-  check_exit_code
-
-  make -j$(nproc) install
-  check_exit_code
+  make -j$(nproc) install || exit
 
   # #################################################
   # install
   # #################################################
   # installing through package manager, which makes uninstalling easy
   if [[ "${install_package}" == true ]]; then
-    make package
-    check_exit_code
+    make package || exit
 
     case "${ID}" in
       ubuntu)
