@@ -1,5 +1,5 @@
 /* ************************************************************************
- * Copyright (C) 2018-2022 Advanced Micro Devices, Inc. All rights Reserved.
+ * Copyright (C) 2018-2023 Advanced Micro Devices, Inc. All rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -114,74 +114,74 @@ namespace rocalution
     }
 
     template <typename ValueType>
-    void HIPAcceleratorMatrixBCSR<ValueType>::AllocateBCSR(int nnzb,
-                                                           int nrowb,
-                                                           int ncolb,
-                                                           int blockdim)
+    void HIPAcceleratorMatrixBCSR<ValueType>::AllocateBCSR(int64_t nnzb,
+                                                           int     nrowb,
+                                                           int     ncolb,
+                                                           int     blockdim)
     {
         assert(nnzb >= 0);
         assert(ncolb >= 0);
         assert(nrowb >= 0);
         assert(blockdim > 1);
 
-        if(this->nnz_ > 0)
-        {
-            this->Clear();
-        }
+        this->Clear();
 
-        if(nnzb > 0)
-        {
-            allocate_hip(nrowb + 1, &this->mat_.row_offset);
-            allocate_hip(nnzb, &this->mat_.col);
-            allocate_hip(nnzb * blockdim * blockdim, &this->mat_.val);
+        allocate_hip(nrowb + 1, &this->mat_.row_offset);
+        allocate_hip(nnzb, &this->mat_.col);
+        allocate_hip(nnzb * blockdim * blockdim, &this->mat_.val);
 
-            set_to_zero_hip(this->local_backend_.HIP_block_size, nrowb + 1, this->mat_.row_offset);
-            set_to_zero_hip(this->local_backend_.HIP_block_size, nnzb, this->mat_.col);
-            set_to_zero_hip(
-                this->local_backend_.HIP_block_size, nnzb * blockdim * blockdim, this->mat_.val);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, nrowb + 1, this->mat_.row_offset);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, nnzb, this->mat_.col);
+        set_to_zero_hip(
+            this->local_backend_.HIP_block_size, nnzb * blockdim * blockdim, this->mat_.val);
 
-            this->nrow_ = nrowb * blockdim;
-            this->ncol_ = ncolb * blockdim;
-            this->nnz_  = nnzb * blockdim * blockdim;
+        this->nrow_ = nrowb * blockdim;
+        this->ncol_ = ncolb * blockdim;
+        this->nnz_  = nnzb * blockdim * blockdim;
 
-            this->mat_.nrowb    = nrowb;
-            this->mat_.ncolb    = ncolb;
-            this->mat_.nnzb     = nnzb;
-            this->mat_.blockdim = blockdim;
-        }
+        this->mat_.nrowb    = nrowb;
+        this->mat_.ncolb    = ncolb;
+        this->mat_.nnzb     = nnzb;
+        this->mat_.blockdim = blockdim;
     }
 
     template <typename ValueType>
     void HIPAcceleratorMatrixBCSR<ValueType>::Clear()
     {
-        if(this->nnz_ > 0)
-        {
-            free_hip(&this->mat_.row_offset);
-            free_hip(&this->mat_.col);
-            free_hip(&this->mat_.val);
+        free_hip(&this->mat_.row_offset);
+        free_hip(&this->mat_.col);
+        free_hip(&this->mat_.val);
 
-            this->nrow_ = 0;
-            this->ncol_ = 0;
-            this->nnz_  = 0;
+        this->nrow_ = 0;
+        this->ncol_ = 0;
+        this->nnz_  = 0;
 
-            this->LAnalyseClear();
-            this->UAnalyseClear();
-            this->LUAnalyseClear();
-            this->LLAnalyseClear();
-        }
+        this->LAnalyseClear();
+        this->UAnalyseClear();
+        this->LUAnalyseClear();
+        this->LLAnalyseClear();
     }
 
     template <typename ValueType>
-    void HIPAcceleratorMatrixBCSR<ValueType>::SetDataPtrBCSR(
-        int** row_offset, int** col, ValueType** val, int nnzb, int nrowb, int ncolb, int blockdim)
+    void HIPAcceleratorMatrixBCSR<ValueType>::SetDataPtrBCSR(int**       row_offset,
+                                                             int**       col,
+                                                             ValueType** val,
+                                                             int64_t     nnzb,
+                                                             int         nrowb,
+                                                             int         ncolb,
+                                                             int         blockdim)
     {
-        assert(*row_offset != NULL);
-        assert(*col != NULL);
-        assert(*val != NULL);
-        assert(nnzb > 0);
-        assert(nrowb > 0);
-        assert(ncolb > 0);
+        assert(nnzb >= 0);
+        assert(nrowb >= 0);
+        assert(ncolb >= 0);
         assert(blockdim > 1);
+        assert(*row_offset != NULL);
+
+        if(nnzb > 0)
+        {
+            assert(*col != NULL);
+            assert(*val != NULL);
+        }
 
         this->Clear();
 
@@ -207,9 +207,9 @@ namespace rocalution
                                                                ValueType** val,
                                                                int&        blockdim)
     {
-        assert(this->nrow_ > 0);
-        assert(this->ncol_ > 0);
-        assert(this->nnz_ > 0);
+        assert(this->nrow_ >= 0);
+        assert(this->ncol_ >= 0);
+        assert(this->nnz_ >= 0);
         assert(this->mat_.blockdim > 1);
 
         hipDeviceSynchronize();
@@ -238,7 +238,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixBCSR<ValueType>*>(&src)) != NULL)
@@ -259,24 +258,16 @@ namespace rocalution
             assert(this->mat_.nnzb == cast_mat->mat_.nnzb);
             assert(this->mat_.blockdim == cast_mat->mat_.blockdim);
 
-            hipMemcpy(this->mat_.row_offset,
-                      cast_mat->mat_.row_offset,
-                      (this->mat_.nrowb + 1) * sizeof(int),
-                      hipMemcpyHostToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
+            // Copy only if initialized
+            if(cast_mat->mat_.row_offset != NULL)
+            {
+                copy_h2d(this->mat_.nrowb + 1, cast_mat->mat_.row_offset, this->mat_.row_offset);
+            }
 
-            hipMemcpy(this->mat_.col,
-                      cast_mat->mat_.col,
-                      this->mat_.nnzb * sizeof(int),
-                      hipMemcpyHostToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-            hipMemcpy(this->mat_.val,
-                      cast_mat->mat_.val,
-                      this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim
-                          * sizeof(ValueType),
-                      hipMemcpyHostToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
+            copy_h2d(this->mat_.nnzb, cast_mat->mat_.col, this->mat_.col);
+            copy_h2d(this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim,
+                     cast_mat->mat_.val,
+                     this->mat_.val);
         }
         else
         {
@@ -314,24 +305,15 @@ namespace rocalution
             assert(this->mat_.nnzb == cast_mat->mat_.nnzb);
             assert(this->mat_.blockdim == cast_mat->mat_.blockdim);
 
-            hipMemcpy(cast_mat->mat_.row_offset,
-                      this->mat_.row_offset,
-                      (this->mat_.nrowb + 1) * sizeof(int),
-                      hipMemcpyDeviceToHost);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
+            if(this->mat_.row_offset != NULL)
+            {
+                copy_d2h(this->mat_.nrowb + 1, this->mat_.row_offset, cast_mat->mat_.row_offset);
+            }
 
-            hipMemcpy(cast_mat->mat_.col,
-                      this->mat_.col,
-                      this->mat_.nnzb * sizeof(int),
-                      hipMemcpyDeviceToHost);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-            hipMemcpy(cast_mat->mat_.val,
-                      this->mat_.val,
-                      this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim
-                          * sizeof(ValueType),
-                      hipMemcpyDeviceToHost);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
+            copy_d2h(this->mat_.nnzb, this->mat_.col, cast_mat->mat_.col);
+            copy_d2h(this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim,
+                     this->mat_.val,
+                     cast_mat->mat_.val);
         }
         else
         {
@@ -350,7 +332,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixBCSR<ValueType>*>(&src)) != NULL)
@@ -371,24 +352,17 @@ namespace rocalution
             assert(this->mat_.nnzb == hip_cast_mat->mat_.nnzb);
             assert(this->mat_.blockdim == hip_cast_mat->mat_.blockdim);
 
-            hipMemcpy(this->mat_.row_offset,
-                      hip_cast_mat->mat_.row_offset,
-                      (this->mat_.nrowb + 1) * sizeof(int),
-                      hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
+            // Copy only if initialized
+            if(hip_cast_mat->mat_.row_offset != NULL)
+            {
+                copy_h2d(
+                    this->mat_.nrowb + 1, hip_cast_mat->mat_.row_offset, this->mat_.row_offset);
+            }
 
-            hipMemcpy(this->mat_.col,
-                      hip_cast_mat->mat_.col,
-                      this->mat_.nnzb * sizeof(int),
-                      hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-            hipMemcpy(this->mat_.val,
-                      hip_cast_mat->mat_.val,
-                      this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim
-                          * sizeof(ValueType),
-                      hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
+            copy_h2d(this->mat_.nnzb, hip_cast_mat->mat_.col, this->mat_.col);
+            copy_h2d(this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim,
+                     hip_cast_mat->mat_.val,
+                     this->mat_.val);
         }
         else
         {
@@ -435,24 +409,16 @@ namespace rocalution
             assert(this->mat_.nnzb == hip_cast_mat->mat_.nnzb);
             assert(this->mat_.blockdim == hip_cast_mat->mat_.blockdim);
 
-            hipMemcpy(hip_cast_mat->mat_.row_offset,
-                      this->mat_.row_offset,
-                      (this->mat_.nrowb + 1) * sizeof(int),
-                      hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
+            if(this->mat_.row_offset != NULL)
+            {
+                copy_d2d(
+                    this->mat_.nrowb + 1, this->mat_.row_offset, hip_cast_mat->mat_.row_offset);
+            }
 
-            hipMemcpy(hip_cast_mat->mat_.col,
-                      this->mat_.col,
-                      this->mat_.nnzb * sizeof(int),
-                      hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-            hipMemcpy(hip_cast_mat->mat_.val,
-                      this->mat_.val,
-                      this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim
-                          * sizeof(ValueType),
-                      hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
+            copy_d2d(this->mat_.nnzb, this->mat_.col, hip_cast_mat->mat_.col);
+            copy_d2d(this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim,
+                     this->mat_.val,
+                     hip_cast_mat->mat_.val);
         }
         else
         {
@@ -478,7 +444,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixBCSR<ValueType>*>(&src)) != NULL)
@@ -499,27 +464,26 @@ namespace rocalution
             assert(this->mat_.nnzb == cast_mat->mat_.nnzb);
             assert(this->mat_.blockdim == cast_mat->mat_.blockdim);
 
-            if(this->nnz_ > 0)
+            // Copy only if initialized
+            if(cast_mat->mat_.row_offset != NULL)
             {
-                hipMemcpyAsync(this->mat_.row_offset,
-                               cast_mat->mat_.row_offset,
-                               (this->mat_.nrowb + 1) * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.col,
-                               cast_mat->mat_.col,
-                               this->mat_.nnzb * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.val,
-                               cast_mat->mat_.val,
-                               this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim
-                                   * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
+                copy_h2d(this->mat_.nrowb + 1,
+                         cast_mat->mat_.row_offset,
+                         this->mat_.row_offset,
+                         true,
+                         HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
             }
+
+            copy_h2d(this->mat_.nnzb,
+                     cast_mat->mat_.col,
+                     this->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_h2d(this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim,
+                     cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -557,27 +521,25 @@ namespace rocalution
             assert(this->mat_.nnzb == cast_mat->mat_.nnzb);
             assert(this->mat_.blockdim == cast_mat->mat_.blockdim);
 
-            if(this->nnz_ > 0)
+            if(this->mat_.row_offset != NULL)
             {
-                hipMemcpyAsync(cast_mat->mat_.row_offset,
-                               this->mat_.row_offset,
-                               (this->mat_.nrowb + 1) * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(cast_mat->mat_.col,
-                               this->mat_.col,
-                               this->mat_.nnzb * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(cast_mat->mat_.val,
-                               this->mat_.val,
-                               this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim
-                                   * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
+                copy_d2h(this->mat_.nrowb + 1,
+                         this->mat_.row_offset,
+                         cast_mat->mat_.row_offset,
+                         true,
+                         HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
             }
+
+            copy_d2h(this->mat_.nnzb,
+                     this->mat_.col,
+                     cast_mat->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim,
+                     this->mat_.val,
+                     cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -596,7 +558,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixBCSR<ValueType>*>(&src)) != NULL)
@@ -617,27 +578,26 @@ namespace rocalution
             assert(this->mat_.nnzb == hip_cast_mat->mat_.nnzb);
             assert(this->mat_.blockdim == hip_cast_mat->mat_.blockdim);
 
-            if(this->nnz_ > 0)
+            // Copy only if initialized
+            if(hip_cast_mat->mat_.row_offset != NULL)
             {
-                hipMemcpyAsync(this->mat_.row_offset,
-                               hip_cast_mat->mat_.row_offset,
-                               (this->mat_.nrowb + 1) * sizeof(int),
-                               hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.col,
-                               hip_cast_mat->mat_.col,
-                               this->mat_.nnzb * sizeof(int),
-                               hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.val,
-                               hip_cast_mat->mat_.val,
-                               this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim
-                                   * sizeof(int),
-                               hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
+                copy_d2d(this->mat_.nrowb + 1,
+                         hip_cast_mat->mat_.row_offset,
+                         this->mat_.row_offset,
+                         true,
+                         HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
             }
+
+            copy_d2d(this->mat_.nnzb,
+                     hip_cast_mat->mat_.col,
+                     this->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim,
+                     hip_cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -684,27 +644,25 @@ namespace rocalution
             assert(this->mat_.nnzb == hip_cast_mat->mat_.nnzb);
             assert(this->mat_.blockdim == hip_cast_mat->mat_.blockdim);
 
-            if(this->nnz_ > 0)
+            if(this->mat_.row_offset != NULL)
             {
-                hipMemcpyAsync(hip_cast_mat->mat_.row_offset,
-                               this->mat_.row_offset,
-                               (this->mat_.nrowb + 1) * sizeof(int),
-                               hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(hip_cast_mat->mat_.col,
-                               this->mat_.col,
-                               this->mat_.nnzb * sizeof(int),
-                               hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(hip_cast_mat->mat_.val,
-                               this->mat_.val,
-                               this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim
-                                   * sizeof(int),
-                               hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
+                copy_d2d(this->mat_.nrowb + 1,
+                         this->mat_.row_offset,
+                         hip_cast_mat->mat_.row_offset,
+                         true,
+                         HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
             }
+
+            copy_d2d(this->mat_.nnzb,
+                     this->mat_.col,
+                     hip_cast_mat->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->mat_.nnzb * this->mat_.blockdim * this->mat_.blockdim,
+                     this->mat_.val,
+                     hip_cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -728,9 +686,14 @@ namespace rocalution
     {
         this->Clear();
 
-        // empty matrix is empty matrix
+        // Empty matrix
         if(mat.GetNnz() == 0)
         {
+            int mb = (mat.GetM() + 1) / 2;
+            int nb = (mat.GetN() + 1) / 2;
+
+            this->AllocateBCSR(0, mb, nb, 2);
+
             return true;
         }
 
@@ -747,7 +710,7 @@ namespace rocalution
         {
             this->Clear();
 
-            if(csr_to_bcsr_hip(ROCSPARSE_HANDLE(this->local_backend_.ROC_sparse_handle),
+            if(csr_to_bcsr_hip(&this->local_backend_,
                                cast_mat_csr->nnz_,
                                cast_mat_csr->nrow_,
                                cast_mat_csr->ncol_,
@@ -798,7 +761,7 @@ namespace rocalution
             if(this->mat_buffer_ == NULL)
             {
                 this->mat_buffer_size_ = buffer_size;
-                hipMalloc(&this->mat_buffer_, buffer_size);
+                allocate_hip(buffer_size, &this->mat_buffer_);
             }
 
             assert(this->mat_buffer_size_ >= buffer_size);
@@ -910,7 +873,7 @@ namespace rocalution
         if(this->mat_buffer_ == NULL)
         {
             this->mat_buffer_size_ = buffer_size;
-            hipMalloc(&this->mat_buffer_, buffer_size);
+            allocate_hip(buffer_size, &this->mat_buffer_);
         }
 
         assert(this->mat_buffer_size_ >= buffer_size);
@@ -993,7 +956,7 @@ namespace rocalution
         // Clear buffer
         if(this->mat_buffer_ != NULL)
         {
-            hipFree(this->mat_buffer_);
+            free_hip(&this->mat_buffer_);
             this->mat_buffer_ = NULL;
         }
 
@@ -1154,13 +1117,13 @@ namespace rocalution
         if(this->mat_buffer_ == NULL)
         {
             this->mat_buffer_size_ = buffer_size;
-            hipMalloc(&this->mat_buffer_, buffer_size);
+            allocate_hip(buffer_size, &this->mat_buffer_);
         }
         else if(this->mat_buffer_size_ < buffer_size)
         {
             this->mat_buffer_size_ = buffer_size;
-            hipFree(this->mat_buffer_);
-            hipMalloc(&this->mat_buffer_, buffer_size);
+            free_hip(&this->mat_buffer_);
+            allocate_hip(buffer_size, &this->mat_buffer_);
         }
 
         assert(this->mat_buffer_size_ >= buffer_size);
@@ -1229,7 +1192,7 @@ namespace rocalution
         // Clear buffer
         if(this->mat_buffer_ != NULL)
         {
-            hipFree(this->mat_buffer_);
+            free_hip(&this->mat_buffer_);
             this->mat_buffer_ = NULL;
         }
 
@@ -1380,7 +1343,7 @@ namespace rocalution
         if(this->mat_buffer_ == NULL)
         {
             this->mat_buffer_size_ = buffer_size;
-            hipMalloc(&this->mat_buffer_, buffer_size);
+            allocate_hip(buffer_size, &this->mat_buffer_);
         }
 
         assert(this->mat_buffer_size_ >= buffer_size);
@@ -1456,7 +1419,7 @@ namespace rocalution
         if(this->mat_buffer_ == NULL)
         {
             this->mat_buffer_size_ = buffer_size;
-            hipMalloc(&this->mat_buffer_, buffer_size);
+            allocate_hip(buffer_size, &this->mat_buffer_);
         }
 
         assert(this->mat_buffer_size_ >= buffer_size);
@@ -1495,7 +1458,7 @@ namespace rocalution
         // Clear buffer
         if(this->mat_buffer_ != NULL)
         {
-            hipFree(this->mat_buffer_);
+            free_hip(&this->mat_buffer_);
             this->mat_buffer_ = NULL;
         }
 
@@ -1527,7 +1490,7 @@ namespace rocalution
         // Clear buffer
         if(this->mat_buffer_ != NULL)
         {
-            hipFree(this->mat_buffer_);
+            free_hip(&this->mat_buffer_);
             this->mat_buffer_ = NULL;
         }
 
