@@ -1,5 +1,5 @@
 /* ************************************************************************
- * Copyright (C) 2018-2022 Advanced Micro Devices, Inc. All rights Reserved.
+ * Copyright (C) 2018-2023 Advanced Micro Devices, Inc. All rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -102,43 +102,41 @@ namespace rocalution
     }
 
     template <typename ValueType>
-    void HIPAcceleratorMatrixCOO<ValueType>::AllocateCOO(int nnz, int nrow, int ncol)
+    void HIPAcceleratorMatrixCOO<ValueType>::AllocateCOO(int64_t nnz, int nrow, int ncol)
     {
         assert(nnz >= 0);
         assert(ncol >= 0);
         assert(nrow >= 0);
 
-        if(this->nnz_ > 0)
-        {
-            this->Clear();
-        }
+        this->Clear();
 
-        if(nnz > 0)
-        {
-            allocate_hip(nnz, &this->mat_.row);
-            allocate_hip(nnz, &this->mat_.col);
-            allocate_hip(nnz, &this->mat_.val);
+        allocate_hip(nnz, &this->mat_.row);
+        allocate_hip(nnz, &this->mat_.col);
+        allocate_hip(nnz, &this->mat_.val);
 
-            set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.row);
-            set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.col);
-            set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.val);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.row);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.col);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.val);
 
-            this->nrow_ = nrow;
-            this->ncol_ = ncol;
-            this->nnz_  = nnz;
-        }
+        this->nrow_ = nrow;
+        this->ncol_ = ncol;
+        this->nnz_  = nnz;
     }
 
     template <typename ValueType>
     void HIPAcceleratorMatrixCOO<ValueType>::SetDataPtrCOO(
-        int** row, int** col, ValueType** val, int nnz, int nrow, int ncol)
+        int** row, int** col, ValueType** val, int64_t nnz, int nrow, int ncol)
     {
-        assert(*row != NULL);
-        assert(*col != NULL);
-        assert(*val != NULL);
-        assert(nnz > 0);
-        assert(nrow > 0);
-        assert(ncol > 0);
+        assert(nnz >= 0);
+        assert(nrow >= 0);
+        assert(ncol >= 0);
+
+        if(nnz > 0)
+        {
+            assert(*row != NULL);
+            assert(*col != NULL);
+            assert(*val != NULL);
+        }
 
         this->Clear();
 
@@ -156,9 +154,9 @@ namespace rocalution
     template <typename ValueType>
     void HIPAcceleratorMatrixCOO<ValueType>::LeaveDataPtrCOO(int** row, int** col, ValueType** val)
     {
-        assert(this->nrow_ > 0);
-        assert(this->ncol_ > 0);
-        assert(this->nnz_ > 0);
+        assert(this->nrow_ >= 0);
+        assert(this->ncol_ >= 0);
+        assert(this->nnz_ >= 0);
 
         hipDeviceSynchronize();
 
@@ -179,16 +177,13 @@ namespace rocalution
     template <typename ValueType>
     void HIPAcceleratorMatrixCOO<ValueType>::Clear()
     {
-        if(this->nnz_ > 0)
-        {
-            free_hip(&this->mat_.row);
-            free_hip(&this->mat_.col);
-            free_hip(&this->mat_.val);
+        free_hip(&this->mat_.row);
+        free_hip(&this->mat_.col);
+        free_hip(&this->mat_.val);
 
-            this->nrow_ = 0;
-            this->ncol_ = 0;
-            this->nnz_  = 0;
-        }
+        this->nrow_ = 0;
+        this->ncol_ = 0;
+        this->nnz_  = 0;
     }
 
     template <typename ValueType>
@@ -198,7 +193,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixCOO<ValueType>*>(&src)) != NULL)
@@ -208,30 +202,13 @@ namespace rocalution
                 this->AllocateCOO(cast_mat->nnz_, cast_mat->nrow_, cast_mat->ncol_);
             }
 
-            if(this->nnz_ > 0)
-            {
-                assert(this->nnz_ == cast_mat->nnz_);
-                assert(this->nrow_ == cast_mat->nrow_);
-                assert(this->ncol_ == cast_mat->ncol_);
+            assert(this->nnz_ == cast_mat->nnz_);
+            assert(this->nrow_ == cast_mat->nrow_);
+            assert(this->ncol_ == cast_mat->ncol_);
 
-                hipMemcpy(this->mat_.row,
-                          cast_mat->mat_.row,
-                          (this->nnz_) * sizeof(int),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.col,
-                          cast_mat->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.val,
-                          cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->nnz_, cast_mat->mat_.row, this->mat_.row);
+            copy_h2d(this->nnz_, cast_mat->mat_.col, this->mat_.col);
+            copy_h2d(this->nnz_, cast_mat->mat_.val, this->mat_.val);
         }
         else
         {
@@ -264,26 +241,9 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(cast_mat->mat_.row,
-                          this->mat_.row,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(cast_mat->mat_.col,
-                          this->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->nnz_, this->mat_.row, cast_mat->mat_.row);
+            copy_d2h(this->nnz_, this->mat_.col, cast_mat->mat_.col);
+            copy_d2h(this->nnz_, this->mat_.val, cast_mat->mat_.val);
         }
         else
         {
@@ -302,7 +262,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixCOO<ValueType>*>(&src)) != NULL)
@@ -316,26 +275,9 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.row,
-                          hip_cast_mat->mat_.row,
-                          (this->nnz_) * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.col,
-                          hip_cast_mat->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.val,
-                          hip_cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_, hip_cast_mat->mat_.row, this->mat_.row);
+            copy_d2d(this->nnz_, hip_cast_mat->mat_.col, this->mat_.col);
+            copy_d2d(this->nnz_, hip_cast_mat->mat_.val, this->mat_.val);
         }
         else
         {
@@ -377,26 +319,9 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(hip_cast_mat->mat_.row,
-                          this->mat_.row,
-                          (this->nnz_) * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.col,
-                          this->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_, this->mat_.row, hip_cast_mat->mat_.row);
+            copy_d2d(this->nnz_, this->mat_.col, hip_cast_mat->mat_.col);
+            copy_d2d(this->nnz_, this->mat_.val, hip_cast_mat->mat_.val);
         }
         else
         {
@@ -422,7 +347,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixCOO<ValueType>*>(&src)) != NULL)
@@ -432,30 +356,25 @@ namespace rocalution
                 this->AllocateCOO(cast_mat->nnz_, cast_mat->nrow_, cast_mat->ncol_);
             }
 
-            if(this->nnz_ > 0)
-            {
-                assert(this->nnz_ == cast_mat->nnz_);
-                assert(this->nrow_ == cast_mat->nrow_);
-                assert(this->ncol_ == cast_mat->ncol_);
+            assert(this->nnz_ == cast_mat->nnz_);
+            assert(this->nrow_ == cast_mat->nrow_);
+            assert(this->ncol_ == cast_mat->ncol_);
 
-                hipMemcpyAsync(this->mat_.row,
-                               cast_mat->mat_.row,
-                               (this->nnz_) * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.col,
-                               cast_mat->mat_.col,
-                               this->nnz_ * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.val,
-                               cast_mat->mat_.val,
-                               this->nnz_ * sizeof(ValueType),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->nnz_,
+                     cast_mat->mat_.row,
+                     this->mat_.row,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_h2d(this->nnz_,
+                     cast_mat->mat_.col,
+                     this->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_h2d(this->nnz_,
+                     cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -484,30 +403,25 @@ namespace rocalution
                 cast_mat->AllocateCOO(this->nnz_, this->nrow_, this->ncol_);
             }
 
-            if(this->nnz_ > 0)
-            {
-                assert(this->nnz_ == cast_mat->nnz_);
-                assert(this->nrow_ == cast_mat->nrow_);
-                assert(this->ncol_ == cast_mat->ncol_);
+            assert(this->nnz_ == cast_mat->nnz_);
+            assert(this->nrow_ == cast_mat->nrow_);
+            assert(this->ncol_ == cast_mat->ncol_);
 
-                hipMemcpyAsync(cast_mat->mat_.row,
-                               this->mat_.row,
-                               this->nnz_ * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(cast_mat->mat_.col,
-                               this->mat_.col,
-                               this->nnz_ * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(cast_mat->mat_.val,
-                               this->mat_.val,
-                               this->nnz_ * sizeof(ValueType),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->nnz_,
+                     this->mat_.row,
+                     cast_mat->mat_.row,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->nnz_,
+                     this->mat_.col,
+                     cast_mat->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->nnz_,
+                     this->mat_.val,
+                     cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -526,7 +440,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixCOO<ValueType>*>(&src)) != NULL)
@@ -540,26 +453,21 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.row,
-                          hip_cast_mat->mat_.row,
-                          (this->nnz_) * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.col,
-                          hip_cast_mat->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.val,
-                          hip_cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_,
+                     hip_cast_mat->mat_.row,
+                     this->mat_.row,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->nnz_,
+                     hip_cast_mat->mat_.col,
+                     this->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->nnz_,
+                     hip_cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -601,26 +509,21 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(hip_cast_mat->mat_.row,
-                          this->mat_.row,
-                          (this->nnz_) * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.col,
-                          this->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->nnz_,
+                     this->mat_.row,
+                     hip_cast_mat->mat_.row,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->nnz_,
+                     this->mat_.col,
+                     hip_cast_mat->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->nnz_,
+                     this->mat_.val,
+                     hip_cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -644,45 +547,17 @@ namespace rocalution
                                                          const int*       col,
                                                          const ValueType* val)
     {
-        // assert CSR format
-        assert(this->GetMatFormat() == COO);
-
-        if(this->nnz_ > 0)
-        {
-            assert(this->nrow_ > 0);
-            assert(this->ncol_ > 0);
-
-            hipMemcpy(this->mat_.row, row, this->nnz_ * sizeof(int), hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-            hipMemcpy(this->mat_.col, col, this->nnz_ * sizeof(int), hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-            hipMemcpy(this->mat_.val, val, this->nnz_ * sizeof(ValueType), hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-        }
+        copy_d2d(this->nnz_, row, this->mat_.row);
+        copy_d2d(this->nnz_, col, this->mat_.col);
+        copy_d2d(this->nnz_, val, this->mat_.val);
     }
 
     template <typename ValueType>
     void HIPAcceleratorMatrixCOO<ValueType>::CopyToCOO(int* row, int* col, ValueType* val) const
     {
-        // assert CSR format
-        assert(this->GetMatFormat() == COO);
-
-        if(this->nnz_ > 0)
-        {
-            assert(this->nrow_ > 0);
-            assert(this->ncol_ > 0);
-
-            hipMemcpy(row, this->mat_.row, this->nnz_ * sizeof(int), hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-            hipMemcpy(col, this->mat_.col, this->nnz_ * sizeof(int), hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-            hipMemcpy(val, this->mat_.val, this->nnz_ * sizeof(ValueType), hipMemcpyDeviceToDevice);
-            CHECK_HIP_ERROR(__FILE__, __LINE__);
-        }
+        copy_d2d(this->nnz_, this->mat_.row, row);
+        copy_d2d(this->nnz_, this->mat_.col, col);
+        copy_d2d(this->nnz_, this->mat_.val, val);
     }
 
     template <typename ValueType>
@@ -693,6 +568,8 @@ namespace rocalution
         // empty matrix is empty matrix
         if(mat.GetNnz() == 0)
         {
+            this->AllocateCOO(mat.GetNnz(), mat.GetM(), mat.GetN());
+
             return true;
         }
 
@@ -710,7 +587,7 @@ namespace rocalution
         {
             this->Clear();
 
-            if(csr_to_coo_hip(ROCSPARSE_HANDLE(this->local_backend_.ROC_sparse_handle),
+            if(csr_to_coo_hip(&this->local_backend_,
                               cast_mat_csr->nnz_,
                               cast_mat_csr->nrow_,
                               cast_mat_csr->ncol_,
@@ -826,9 +703,9 @@ namespace rocalution
             src.AllocateCOO(this->nnz_, this->nrow_, this->ncol_);
             src.CopyFrom(*this);
 
-            int nnz = this->nnz_;
-            int s   = nnz;
-            int k
+            int64_t nnz = this->nnz_;
+            int64_t s   = nnz;
+            int64_t k
                 = (nnz / this->local_backend_.HIP_block_size) / this->local_backend_.HIP_max_threads
                   + 1;
             if(k > 1)
@@ -839,17 +716,16 @@ namespace rocalution
             dim3 BlockSize(this->local_backend_.HIP_block_size);
             dim3 GridSize(s / this->local_backend_.HIP_block_size + 1);
 
-            hipLaunchKernelGGL((kernel_coo_permute<ValueType, int>),
-                               GridSize,
-                               BlockSize,
-                               0,
-                               0,
-                               nnz,
-                               src.mat_.row,
-                               src.mat_.col,
-                               cast_perm->vec_,
-                               this->mat_.row,
-                               this->mat_.col);
+            kernel_coo_permute<ValueType, int>
+                <<<GridSize,
+                   BlockSize,
+                   0,
+                   HIPSTREAM(_get_backend_descriptor()->HIP_stream_current)>>>(nnz,
+                                                                               src.mat_.row,
+                                                                               src.mat_.col,
+                                                                               cast_perm->vec_,
+                                                                               this->mat_.row,
+                                                                               this->mat_.col);
             CHECK_HIP_ERROR(__FILE__, __LINE__);
         }
 
@@ -877,17 +753,20 @@ namespace rocalution
             dim3 BlockSize1(this->local_backend_.HIP_block_size);
             dim3 GridSize1(n / this->local_backend_.HIP_block_size + 1);
 
-            hipLaunchKernelGGL(
-                (kernel_reverse_index<int>), GridSize1, BlockSize1, 0, 0, n, cast_perm->vec_, pb);
+            kernel_reverse_index<<<GridSize1,
+                                   BlockSize1,
+                                   0,
+                                   HIPSTREAM(_get_backend_descriptor()->HIP_stream_current)>>>(
+                n, cast_perm->vec_, pb);
             CHECK_HIP_ERROR(__FILE__, __LINE__);
 
             HIPAcceleratorMatrixCOO<ValueType> src(this->local_backend_);
             src.AllocateCOO(this->nnz_, this->nrow_, this->ncol_);
             src.CopyFrom(*this);
 
-            int nnz = this->nnz_;
-            int s   = nnz;
-            int k
+            int64_t nnz = this->nnz_;
+            int64_t s   = nnz;
+            int64_t k
                 = (nnz / this->local_backend_.HIP_block_size) / this->local_backend_.HIP_max_threads
                   + 1;
             if(k > 1)
@@ -898,17 +777,12 @@ namespace rocalution
             dim3 BlockSize2(this->local_backend_.HIP_block_size);
             dim3 GridSize2(s / this->local_backend_.HIP_block_size + 1);
 
-            hipLaunchKernelGGL((kernel_coo_permute<ValueType, int>),
-                               GridSize2,
-                               BlockSize2,
-                               0,
-                               0,
-                               nnz,
-                               src.mat_.row,
-                               src.mat_.col,
-                               pb,
-                               this->mat_.row,
-                               this->mat_.col);
+            kernel_coo_permute<ValueType, int>
+                <<<GridSize2,
+                   BlockSize2,
+                   0,
+                   HIPSTREAM(_get_backend_descriptor()->HIP_stream_current)>>>(
+                    nnz, src.mat_.row, src.mat_.col, pb, this->mat_.row, this->mat_.col);
             CHECK_HIP_ERROR(__FILE__, __LINE__);
 
             free_hip(&pb);
@@ -935,8 +809,8 @@ namespace rocalution
                 &buffer_size);
             CHECK_ROCSPARSE_ERROR(status, __FILE__, __LINE__);
 
-            void* buffer = nullptr;
-            hipMalloc(&buffer, buffer_size);
+            char* buffer = nullptr;
+            allocate_hip(buffer_size, &buffer);
 
             int* perm = nullptr;
             allocate_hip(this->nnz_, &perm);
@@ -970,7 +844,7 @@ namespace rocalution
 
             this->mat_.val = coo_val_sorted;
 
-            hipFree(buffer);
+            free_hip(&buffer);
         }
 
         return true;
