@@ -1,5 +1,5 @@
 /* ************************************************************************
- * Copyright (C) 2018-2022 Advanced Micro Devices, Inc. All rights Reserved.
+ * Copyright (C) 2018-2023 Advanced Micro Devices, Inc. All rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -92,41 +92,37 @@ namespace rocalution
         assert(ncol >= 0);
         assert(nrow >= 0);
 
-        if(this->nnz_ > 0)
-        {
-            this->Clear();
-        }
+        this->Clear();
 
-        if(nrow * ncol > 0)
-        {
-            allocate_hip(nrow * ncol, &this->mat_.val);
-            set_to_zero_hip(this->local_backend_.HIP_block_size, nrow * ncol, mat_.val);
+        allocate_hip(nrow * ncol, &this->mat_.val);
 
-            this->nrow_ = nrow;
-            this->ncol_ = ncol;
-            this->nnz_  = nrow * ncol;
-        }
+        set_to_zero_hip(this->local_backend_.HIP_block_size, nrow * ncol, mat_.val);
+
+        this->nrow_ = nrow;
+        this->ncol_ = ncol;
+        this->nnz_  = nrow * ncol;
     }
 
     template <typename ValueType>
     void HIPAcceleratorMatrixDENSE<ValueType>::Clear()
     {
-        if(this->nnz_ > 0)
-        {
-            free_hip(&this->mat_.val);
+        free_hip(&this->mat_.val);
 
-            this->nrow_ = 0;
-            this->ncol_ = 0;
-            this->nnz_  = 0;
-        }
+        this->nrow_ = 0;
+        this->ncol_ = 0;
+        this->nnz_  = 0;
     }
 
     template <typename ValueType>
     void HIPAcceleratorMatrixDENSE<ValueType>::SetDataPtrDENSE(ValueType** val, int nrow, int ncol)
     {
-        assert(*val != NULL);
-        assert(nrow > 0);
-        assert(ncol > 0);
+        assert(nrow >= 0);
+        assert(ncol >= 0);
+
+        if(nrow * ncol > 0)
+        {
+            assert(*val != NULL);
+        }
 
         this->Clear();
 
@@ -142,9 +138,9 @@ namespace rocalution
     template <typename ValueType>
     void HIPAcceleratorMatrixDENSE<ValueType>::LeaveDataPtrDENSE(ValueType** val)
     {
-        assert(this->nrow_ > 0);
-        assert(this->ncol_ > 0);
-        assert(this->nnz_ > 0);
+        assert(this->nrow_ >= 0);
+        assert(this->ncol_ >= 0);
+        assert(this->nnz_ >= 0);
         assert(this->nnz_ == this->nrow_ * this->ncol_);
 
         hipDeviceSynchronize();
@@ -165,7 +161,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixDENSE<ValueType>*>(&src)) != NULL)
@@ -179,14 +174,7 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.val,
-                          cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->nnz_, cast_mat->mat_.val, this->mat_.val);
         }
         else
         {
@@ -219,14 +207,7 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->nnz_, this->mat_.val, cast_mat->mat_.val);
         }
         else
         {
@@ -245,7 +226,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixDENSE<ValueType>*>(&src)) != NULL)
@@ -259,14 +239,7 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.val,
-                          hip_cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_, hip_cast_mat->mat_.val, this->mat_.val);
         }
         else
         {
@@ -308,14 +281,7 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(hip_cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_, this->mat_.val, hip_cast_mat->mat_.val);
         }
         else
         {
@@ -341,7 +307,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixDENSE<ValueType>*>(&src)) != NULL)
@@ -355,14 +320,11 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpyAsync(this->mat_.val,
-                               cast_mat->mat_.val,
-                               this->nnz_ * sizeof(ValueType),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->nnz_,
+                     cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -395,14 +357,11 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpyAsync(cast_mat->mat_.val,
-                               this->mat_.val,
-                               this->nnz_ * sizeof(ValueType),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->nnz_,
+                     this->mat_.val,
+                     cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -421,7 +380,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixDENSE<ValueType>*>(&src)) != NULL)
@@ -435,14 +393,11 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.val,
-                          hip_cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_,
+                     hip_cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -484,14 +439,11 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(hip_cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->nnz_,
+                     this->mat_.val,
+                     hip_cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -515,9 +467,11 @@ namespace rocalution
     {
         this->Clear();
 
-        // empty matrix is empty matrix
+        // Empty matrix
         if(mat.GetNnz() == 0)
         {
+            this->AllocateDENSE(mat.GetM(), mat.GetN());
+
             return true;
         }
 
@@ -535,8 +489,7 @@ namespace rocalution
         {
             this->Clear();
 
-            if(csr_to_dense_hip(ROCSPARSE_HANDLE(this->local_backend_.ROC_sparse_handle),
-                                ROCBLAS_HANDLE(this->local_backend_.ROC_blas_handle),
+            if(csr_to_dense_hip(&this->local_backend_,
                                 cast_mat_csr->nrow_,
                                 cast_mat_csr->ncol_,
                                 cast_mat_csr->mat_,
@@ -748,16 +701,12 @@ namespace rocalution
             dim3 BlockSize(this->local_backend_.HIP_block_size);
             dim3 GridSize(this->nrow_ / this->local_backend_.HIP_block_size + 1);
 
-            hipLaunchKernelGGL((kernel_dense_replace_column_vector<ValueType, int>),
-                               GridSize,
-                               BlockSize,
-                               0,
-                               0,
-                               cast_vec->vec_,
-                               idx,
-                               this->nrow_,
-                               this->ncol_,
-                               this->mat_.val);
+            kernel_dense_replace_column_vector<<<
+                GridSize,
+                BlockSize,
+                0,
+                HIPSTREAM(_get_backend_descriptor()->HIP_stream_current)>>>(
+                cast_vec->vec_, idx, this->nrow_, this->ncol_, this->mat_.val);
             CHECK_HIP_ERROR(__FILE__, __LINE__);
         }
 
@@ -779,16 +728,12 @@ namespace rocalution
             dim3 BlockSize(this->local_backend_.HIP_block_size);
             dim3 GridSize(this->ncol_ / this->local_backend_.HIP_block_size + 1);
 
-            hipLaunchKernelGGL((kernel_dense_replace_row_vector<ValueType, int>),
-                               GridSize,
-                               BlockSize,
-                               0,
-                               0,
-                               cast_vec->vec_,
-                               idx,
-                               this->nrow_,
-                               this->ncol_,
-                               this->mat_.val);
+            kernel_dense_replace_row_vector<<<GridSize,
+                                              BlockSize,
+                                              0,
+                                              HIPSTREAM(
+                                                  _get_backend_descriptor()->HIP_stream_current)>>>(
+                cast_vec->vec_, idx, this->nrow_, this->ncol_, this->mat_.val);
             CHECK_HIP_ERROR(__FILE__, __LINE__);
         }
 
@@ -811,16 +756,12 @@ namespace rocalution
             dim3 BlockSize(this->local_backend_.HIP_block_size);
             dim3 GridSize(this->nrow_ / this->local_backend_.HIP_block_size + 1);
 
-            hipLaunchKernelGGL((kernel_dense_extract_column_vector<ValueType, int>),
-                               GridSize,
-                               BlockSize,
-                               0,
-                               0,
-                               cast_vec->vec_,
-                               idx,
-                               this->nrow_,
-                               this->ncol_,
-                               this->mat_.val);
+            kernel_dense_extract_column_vector<<<
+                GridSize,
+                BlockSize,
+                0,
+                HIPSTREAM(_get_backend_descriptor()->HIP_stream_current)>>>(
+                cast_vec->vec_, idx, this->nrow_, this->ncol_, this->mat_.val);
             CHECK_HIP_ERROR(__FILE__, __LINE__);
         }
 
@@ -843,16 +784,12 @@ namespace rocalution
             dim3 BlockSize(this->local_backend_.HIP_block_size);
             dim3 GridSize(this->ncol_ / this->local_backend_.HIP_block_size + 1);
 
-            hipLaunchKernelGGL((kernel_dense_extract_row_vector<ValueType, int>),
-                               GridSize,
-                               BlockSize,
-                               0,
-                               0,
-                               cast_vec->vec_,
-                               idx,
-                               this->nrow_,
-                               this->ncol_,
-                               this->mat_.val);
+            kernel_dense_extract_row_vector<<<GridSize,
+                                              BlockSize,
+                                              0,
+                                              HIPSTREAM(
+                                                  _get_backend_descriptor()->HIP_stream_current)>>>(
+                cast_vec->vec_, idx, this->nrow_, this->ncol_, this->mat_.val);
             CHECK_HIP_ERROR(__FILE__, __LINE__);
         }
 
