@@ -1,5 +1,5 @@
 /* ************************************************************************
- * Copyright (C) 2018-2022 Advanced Micro Devices, Inc. All rights Reserved.
+ * Copyright (C) 2018-2023 Advanced Micro Devices, Inc. All rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -84,45 +84,46 @@ namespace rocalution
     }
 
     template <typename ValueType>
-    void HIPAcceleratorMatrixDIA<ValueType>::AllocateDIA(int nnz, int nrow, int ncol, int ndiag)
+    void HIPAcceleratorMatrixDIA<ValueType>::AllocateDIA(int64_t nnz, int nrow, int ncol, int ndiag)
     {
         assert(nnz >= 0);
         assert(ncol >= 0);
         assert(nrow >= 0);
+        assert(ndiag >= 0);
 
-        if(this->nnz_ > 0)
-        {
-            this->Clear();
-        }
+        this->Clear();
 
-        if(nnz > 0)
-        {
-            assert(ndiag > 0);
+        allocate_hip(nnz, &this->mat_.val);
+        allocate_hip(ndiag, &this->mat_.offset);
 
-            allocate_hip(nnz, &this->mat_.val);
-            allocate_hip(ndiag, &this->mat_.offset);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, mat_.val);
 
-            set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, mat_.val);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, ndiag, mat_.offset);
 
-            set_to_zero_hip(this->local_backend_.HIP_block_size, ndiag, mat_.offset);
-
-            this->nrow_         = nrow;
-            this->ncol_         = ncol;
-            this->nnz_          = nnz;
-            this->mat_.num_diag = ndiag;
-        }
+        this->nrow_         = nrow;
+        this->ncol_         = ncol;
+        this->nnz_          = nnz;
+        this->mat_.num_diag = ndiag;
     }
 
     template <typename ValueType>
     void HIPAcceleratorMatrixDIA<ValueType>::SetDataPtrDIA(
-        int** offset, ValueType** val, int nnz, int nrow, int ncol, int num_diag)
+        int** offset, ValueType** val, int64_t nnz, int nrow, int ncol, int num_diag)
     {
-        assert(*offset != NULL);
-        assert(*val != NULL);
-        assert(nnz > 0);
-        assert(nrow > 0);
-        assert(ncol > 0);
-        assert(num_diag > 0);
+        assert(nnz >= 0);
+        assert(nrow >= 0);
+        assert(ncol >= 0);
+        assert(num_diag >= 0);
+
+        if(num_diag > 0)
+        {
+            assert(*offset != NULL);
+        }
+
+        if(nnz > 0)
+        {
+            assert(*val != NULL);
+        }
 
         if(nrow < ncol)
         {
@@ -151,10 +152,10 @@ namespace rocalution
                                                              ValueType** val,
                                                              int&        num_diag)
     {
-        assert(this->nrow_ > 0);
-        assert(this->ncol_ > 0);
-        assert(this->nnz_ > 0);
-        assert(this->mat_.num_diag > 0);
+        assert(this->nrow_ >= 0);
+        assert(this->ncol_ >= 0);
+        assert(this->nnz_ >= 0);
+        assert(this->mat_.num_diag >= 0);
 
         if(this->nrow_ < this->ncol_)
         {
@@ -167,7 +168,6 @@ namespace rocalution
 
         hipDeviceSynchronize();
 
-        // see free_host function for details
         *offset = this->mat_.offset;
         *val    = this->mat_.val;
 
@@ -185,16 +185,13 @@ namespace rocalution
     template <typename ValueType>
     void HIPAcceleratorMatrixDIA<ValueType>::Clear()
     {
-        if(this->nnz_ > 0)
-        {
-            free_hip(&this->mat_.val);
-            free_hip(&this->mat_.offset);
+        free_hip(&this->mat_.val);
+        free_hip(&this->mat_.offset);
 
-            this->nrow_         = 0;
-            this->ncol_         = 0;
-            this->nnz_          = 0;
-            this->mat_.num_diag = 0;
-        }
+        this->nrow_         = 0;
+        this->ncol_         = 0;
+        this->nnz_          = 0;
+        this->mat_.num_diag = 0;
     }
 
     template <typename ValueType>
@@ -204,7 +201,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixDIA<ValueType>*>(&src)) != NULL)
@@ -219,20 +215,8 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.offset,
-                          cast_mat->mat_.offset,
-                          this->mat_.num_diag * sizeof(int),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.val,
-                          cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->mat_.num_diag, cast_mat->mat_.offset, this->mat_.offset);
+            copy_h2d(this->nnz_, cast_mat->mat_.val, this->mat_.val);
         }
         else
         {
@@ -265,20 +249,8 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(cast_mat->mat_.offset,
-                          this->mat_.offset,
-                          this->mat_.num_diag * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->mat_.num_diag, this->mat_.offset, cast_mat->mat_.offset);
+            copy_d2h(this->nnz_, this->mat_.val, cast_mat->mat_.val);
         }
         else
         {
@@ -297,7 +269,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixDIA<ValueType>*>(&src)) != NULL)
@@ -314,20 +285,8 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.offset,
-                          hip_cast_mat->mat_.offset,
-                          this->mat_.num_diag * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.val,
-                          hip_cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->mat_.num_diag, hip_cast_mat->mat_.offset, this->mat_.offset);
+            copy_d2d(this->nnz_, hip_cast_mat->mat_.val, this->mat_.val);
         }
         else
         {
@@ -370,20 +329,8 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(hip_cast_mat->mat_.offset,
-                          this->mat_.offset,
-                          this->mat_.num_diag * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->mat_.num_diag, this->mat_.offset, hip_cast_mat->mat_.offset);
+            copy_d2d(this->nnz_, this->mat_.val, hip_cast_mat->mat_.val);
         }
         else
         {
@@ -409,7 +356,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixDIA<ValueType>*>(&src)) != NULL)
@@ -424,20 +370,16 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpyAsync(this->mat_.offset,
-                               cast_mat->mat_.offset,
-                               this->mat_.num_diag * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.val,
-                               cast_mat->mat_.val,
-                               this->nnz_ * sizeof(ValueType),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->mat_.num_diag,
+                     cast_mat->mat_.offset,
+                     this->mat_.offset,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_h2d(this->nnz_,
+                     cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -470,20 +412,16 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpyAsync(cast_mat->mat_.offset,
-                               this->mat_.offset,
-                               this->mat_.num_diag * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(cast_mat->mat_.val,
-                               this->mat_.val,
-                               this->nnz_ * sizeof(ValueType),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->mat_.num_diag,
+                     this->mat_.offset,
+                     cast_mat->mat_.offset,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->nnz_,
+                     this->mat_.val,
+                     cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -502,7 +440,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixDIA<ValueType>*>(&src)) != NULL)
@@ -519,20 +456,16 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.offset,
-                          hip_cast_mat->mat_.offset,
-                          this->mat_.num_diag * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.val,
-                          hip_cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->mat_.num_diag,
+                     hip_cast_mat->mat_.offset,
+                     this->mat_.offset,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->nnz_,
+                     hip_cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -575,20 +508,16 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(hip_cast_mat->mat_.offset,
-                          this->mat_.offset,
-                          this->mat_.num_diag * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->mat_.num_diag,
+                     this->mat_.offset,
+                     hip_cast_mat->mat_.offset,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->nnz_,
+                     this->mat_.val,
+                     hip_cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -612,9 +541,11 @@ namespace rocalution
     {
         this->Clear();
 
-        // empty matrix is empty matrix
+        // Empty matrix
         if(mat.GetNnz() == 0)
         {
+            this->AllocateDIA(0, mat.GetM(), mat.GetN(), 0);
+
             return true;
         }
 
@@ -631,10 +562,10 @@ namespace rocalution
         {
             this->Clear();
 
-            int nnz_dia;
-            int num_diag;
+            int64_t nnz_dia;
+            int     num_diag;
 
-            if(csr_to_dia_hip(this->local_backend_.HIP_block_size,
+            if(csr_to_dia_hip(&this->local_backend_,
                               cast_mat_csr->nnz_,
                               cast_mat_csr->nrow_,
                               cast_mat_csr->ncol_,
@@ -681,18 +612,17 @@ namespace rocalution
             dim3 BlockSize(this->local_backend_.HIP_block_size);
             dim3 GridSize(nrow / this->local_backend_.HIP_block_size + 1);
 
-            hipLaunchKernelGGL((kernel_dia_spmv<ValueType, int>),
-                               GridSize,
-                               BlockSize,
-                               0,
-                               0,
-                               nrow,
-                               ncol,
-                               num_diag,
-                               this->mat_.offset,
-                               this->mat_.val,
-                               cast_in->vec_,
-                               cast_out->vec_);
+            kernel_dia_spmv<<<GridSize,
+                              BlockSize,
+                              0,
+                              HIPSTREAM(_get_backend_descriptor()->HIP_stream_current)>>>(
+                nrow,
+                ncol,
+                num_diag,
+                this->mat_.offset,
+                this->mat_.val,
+                cast_in->vec_,
+                cast_out->vec_);
             CHECK_HIP_ERROR(__FILE__, __LINE__);
         }
     }
@@ -723,19 +653,18 @@ namespace rocalution
             dim3 BlockSize(this->local_backend_.HIP_block_size);
             dim3 GridSize(nrow / this->local_backend_.HIP_block_size + 1);
 
-            hipLaunchKernelGGL((kernel_dia_add_spmv<ValueType, int>),
-                               GridSize,
-                               BlockSize,
-                               0,
-                               0,
-                               nrow,
-                               ncol,
-                               num_diag,
-                               this->mat_.offset,
-                               this->mat_.val,
-                               scalar,
-                               cast_in->vec_,
-                               cast_out->vec_);
+            kernel_dia_add_spmv<<<GridSize,
+                                  BlockSize,
+                                  0,
+                                  HIPSTREAM(_get_backend_descriptor()->HIP_stream_current)>>>(
+                nrow,
+                ncol,
+                num_diag,
+                this->mat_.offset,
+                this->mat_.val,
+                scalar,
+                cast_in->vec_,
+                cast_out->vec_);
             CHECK_HIP_ERROR(__FILE__, __LINE__);
         }
     }
