@@ -1,5 +1,5 @@
 /* ************************************************************************
- * Copyright (C) 2018-2022 Advanced Micro Devices, Inc. All rights Reserved.
+ * Copyright (C) 2018-2023 Advanced Micro Devices, Inc. All rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -99,61 +99,62 @@ namespace rocalution
     }
 
     template <typename ValueType>
-    void HIPAcceleratorMatrixELL<ValueType>::AllocateELL(int nnz, int nrow, int ncol, int max_row)
+    void HIPAcceleratorMatrixELL<ValueType>::AllocateELL(int64_t nnz,
+                                                         int     nrow,
+                                                         int     ncol,
+                                                         int     max_row)
     {
         assert(nnz >= 0);
         assert(ncol >= 0);
         assert(nrow >= 0);
         assert(max_row >= 0);
 
-        if(this->nnz_ > 0)
-        {
-            this->Clear();
-        }
+        this->Clear();
 
         if(nnz > 0)
         {
             assert(nnz == max_row * nrow);
-
-            allocate_hip(nnz, &this->mat_.val);
-            allocate_hip(nnz, &this->mat_.col);
-
-            set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.val);
-            set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.col);
-
-            this->mat_.max_row = max_row;
-            this->nrow_        = nrow;
-            this->ncol_        = ncol;
-            this->nnz_         = nnz;
         }
+
+        allocate_hip(nnz, &this->mat_.val);
+        allocate_hip(nnz, &this->mat_.col);
+
+        set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.val);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, nnz, this->mat_.col);
+
+        this->mat_.max_row = max_row;
+        this->nrow_        = nrow;
+        this->ncol_        = ncol;
+        this->nnz_         = nnz;
     }
 
     template <typename ValueType>
     void HIPAcceleratorMatrixELL<ValueType>::Clear()
     {
-        if(this->nnz_ > 0)
-        {
-            free_hip(&this->mat_.val);
-            free_hip(&this->mat_.col);
+        free_hip(&this->mat_.val);
+        free_hip(&this->mat_.col);
 
-            this->nrow_ = 0;
-            this->ncol_ = 0;
-            this->nnz_  = 0;
-        }
+        this->nrow_ = 0;
+        this->ncol_ = 0;
+        this->nnz_  = 0;
     }
 
     template <typename ValueType>
     void HIPAcceleratorMatrixELL<ValueType>::SetDataPtrELL(
-        int** col, ValueType** val, int nnz, int nrow, int ncol, int max_row)
+        int** col, ValueType** val, int64_t nnz, int nrow, int ncol, int max_row)
     {
-        assert(*col != NULL);
-        assert(*val != NULL);
-        assert(nnz > 0);
-        assert(nrow > 0);
-        assert(ncol > 0);
-        assert(max_row > 0);
+        assert(nnz >= 0);
+        assert(nrow >= 0);
+        assert(ncol >= 0);
+        assert(max_row >= 0);
         assert(max_row * nrow == nnz);
 
+        if(nnz > 0)
+        {
+            assert(*col != NULL);
+            assert(*val != NULL);
+        }
+
         this->Clear();
 
         hipDeviceSynchronize();
@@ -172,10 +173,10 @@ namespace rocalution
                                                              ValueType** val,
                                                              int&        max_row)
     {
-        assert(this->nrow_ > 0);
-        assert(this->ncol_ > 0);
-        assert(this->nnz_ > 0);
-        assert(this->mat_.max_row > 0);
+        assert(this->nrow_ >= 0);
+        assert(this->ncol_ >= 0);
+        assert(this->nnz_ >= 0);
+        assert(this->mat_.max_row >= 0);
         assert(this->mat_.max_row * this->nrow_ == this->nnz_);
 
         hipDeviceSynchronize();
@@ -202,7 +203,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixELL<ValueType>*>(&src)) != NULL)
@@ -217,20 +217,8 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.col,
-                          cast_mat->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.val,
-                          cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->nnz_, cast_mat->mat_.col, this->mat_.col);
+            copy_h2d(this->nnz_, cast_mat->mat_.val, this->mat_.val);
         }
         else
         {
@@ -263,20 +251,8 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(cast_mat->mat_.col,
-                          this->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->nnz_, this->mat_.col, cast_mat->mat_.col);
+            copy_d2h(this->nnz_, this->mat_.val, cast_mat->mat_.val);
         }
         else
         {
@@ -295,7 +271,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixELL<ValueType>*>(&src)) != NULL)
@@ -312,20 +287,8 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.col,
-                          hip_cast_mat->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.val,
-                          hip_cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_, hip_cast_mat->mat_.col, this->mat_.col);
+            copy_d2d(this->nnz_, hip_cast_mat->mat_.val, this->mat_.val);
         }
         else
         {
@@ -368,20 +331,8 @@ namespace rocalution
             assert(this->ncol_ == hip_cast_mat->ncol_);
             assert(this->mat_.max_row == hip_cast_mat->mat_.max_row);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(hip_cast_mat->mat_.col,
-                          this->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_, this->mat_.col, hip_cast_mat->mat_.col);
+            copy_d2d(this->nnz_, this->mat_.val, hip_cast_mat->mat_.val);
         }
         else
         {
@@ -407,7 +358,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixELL<ValueType>*>(&src)) != NULL)
@@ -422,20 +372,16 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpyAsync(this->mat_.col,
-                               cast_mat->mat_.col,
-                               this->nnz_ * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.val,
-                               cast_mat->mat_.val,
-                               this->nnz_ * sizeof(ValueType),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->nnz_,
+                     cast_mat->mat_.col,
+                     this->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_h2d(this->nnz_,
+                     cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -468,20 +414,16 @@ namespace rocalution
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpyAsync(cast_mat->mat_.col,
-                               this->mat_.col,
-                               this->nnz_ * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(cast_mat->mat_.val,
-                               this->mat_.val,
-                               this->nnz_ * sizeof(ValueType),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->nnz_,
+                     this->mat_.col,
+                     cast_mat->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->nnz_,
+                     this->mat_.val,
+                     cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -500,7 +442,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixELL<ValueType>*>(&src)) != NULL)
@@ -517,20 +458,16 @@ namespace rocalution
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(this->mat_.col,
-                          hip_cast_mat->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.val,
-                          hip_cast_mat->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_,
+                     hip_cast_mat->mat_.col,
+                     this->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->nnz_,
+                     hip_cast_mat->mat_.val,
+                     this->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -573,20 +510,16 @@ namespace rocalution
             assert(this->ncol_ == hip_cast_mat->ncol_);
             assert(this->mat_.max_row == hip_cast_mat->mat_.max_row);
 
-            if(this->nnz_ > 0)
-            {
-                hipMemcpy(hip_cast_mat->mat_.col,
-                          this->mat_.col,
-                          this->nnz_ * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.val,
-                          this->mat_.val,
-                          this->nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->nnz_,
+                     this->mat_.col,
+                     hip_cast_mat->mat_.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->nnz_,
+                     this->mat_.val,
+                     hip_cast_mat->mat_.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -610,9 +543,10 @@ namespace rocalution
     {
         this->Clear();
 
-        // empty matrix is empty matrix
+        // Empty matrix
         if(mat.GetNnz() == 0)
         {
+            this->AllocateELL(0, mat.GetM(), mat.GetN(), 0);
             return true;
         }
 
@@ -629,9 +563,9 @@ namespace rocalution
         {
             this->Clear();
 
-            int ell_nnz;
+            int64_t ell_nnz;
 
-            if(csr_to_ell_hip(ROCSPARSE_HANDLE(this->local_backend_.ROC_sparse_handle),
+            if(csr_to_ell_hip(&this->local_backend_,
                               cast_mat_csr->nnz_,
                               cast_mat_csr->nrow_,
                               cast_mat_csr->ncol_,
