@@ -1,5 +1,5 @@
 /* ************************************************************************
- * Copyright (C) 2018-2022 Advanced Micro Devices, Inc. All rights Reserved.
+ * Copyright (C) 2018-2023 Advanced Micro Devices, Inc. All rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -126,7 +126,7 @@ namespace rocalution
 
     template <typename ValueType>
     void HIPAcceleratorMatrixHYB<ValueType>::AllocateHYB(
-        int ell_nnz, int coo_nnz, int ell_max_row, int nrow, int ncol)
+        int64_t ell_nnz, int64_t coo_nnz, int ell_max_row, int nrow, int ncol)
     {
         assert(ell_nnz >= 0);
         assert(coo_nnz >= 0);
@@ -135,74 +135,54 @@ namespace rocalution
         assert(ncol >= 0);
         assert(nrow >= 0);
 
-        if(this->nnz_ > 0)
-        {
-            this->Clear();
-        }
+        this->Clear();
 
+        this->nnz_  = 0;
         this->nrow_ = nrow;
         this->ncol_ = ncol;
-        this->nnz_  = 0;
 
-        if(ell_nnz > 0)
-        {
-            // ELL
-            assert(ell_nnz == ell_max_row * nrow);
+        // ELL
+        assert(ell_nnz == ell_max_row * nrow);
 
-            allocate_hip(ell_nnz, &this->mat_.ELL.val);
-            allocate_hip(ell_nnz, &this->mat_.ELL.col);
+        allocate_hip(ell_nnz, &this->mat_.ELL.val);
+        allocate_hip(ell_nnz, &this->mat_.ELL.col);
 
-            set_to_zero_hip(this->local_backend_.HIP_block_size, ell_nnz, this->mat_.ELL.val);
-            set_to_zero_hip(this->local_backend_.HIP_block_size, ell_nnz, this->mat_.ELL.col);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, ell_nnz, this->mat_.ELL.val);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, ell_nnz, this->mat_.ELL.col);
 
-            this->mat_.ELL.max_row = ell_max_row;
-            this->ell_nnz_         = ell_nnz;
-            this->nnz_ += ell_nnz;
-        }
+        this->mat_.ELL.max_row = ell_max_row;
+        this->ell_nnz_         = ell_nnz;
+        this->nnz_ += ell_nnz;
 
-        if(coo_nnz > 0)
-        {
-            // COO
-            allocate_hip(coo_nnz, &this->mat_.COO.row);
-            allocate_hip(coo_nnz, &this->mat_.COO.col);
-            allocate_hip(coo_nnz, &this->mat_.COO.val);
+        // COO
+        allocate_hip(coo_nnz, &this->mat_.COO.row);
+        allocate_hip(coo_nnz, &this->mat_.COO.col);
+        allocate_hip(coo_nnz, &this->mat_.COO.val);
 
-            set_to_zero_hip(this->local_backend_.HIP_block_size, coo_nnz, this->mat_.COO.row);
-            set_to_zero_hip(this->local_backend_.HIP_block_size, coo_nnz, this->mat_.COO.col);
-            set_to_zero_hip(this->local_backend_.HIP_block_size, coo_nnz, this->mat_.COO.val);
-            this->coo_nnz_ = coo_nnz;
+        set_to_zero_hip(this->local_backend_.HIP_block_size, coo_nnz, this->mat_.COO.row);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, coo_nnz, this->mat_.COO.col);
+        set_to_zero_hip(this->local_backend_.HIP_block_size, coo_nnz, this->mat_.COO.val);
+        this->coo_nnz_ = coo_nnz;
 
-            this->nnz_ += coo_nnz;
-        }
+        this->nnz_ += coo_nnz;
     }
 
     template <typename ValueType>
     void HIPAcceleratorMatrixHYB<ValueType>::Clear()
     {
-        if(this->nnz_ > 0)
-        {
-            if(this->ell_nnz_ > 0)
-            {
-                free_hip(&this->mat_.ELL.val);
-                free_hip(&this->mat_.ELL.col);
+        free_hip(&this->mat_.ELL.val);
+        free_hip(&this->mat_.ELL.col);
+        free_hip(&this->mat_.COO.row);
+        free_hip(&this->mat_.COO.col);
+        free_hip(&this->mat_.COO.val);
 
-                this->ell_nnz_         = 0;
-                this->mat_.ELL.max_row = 0;
-            }
+        this->ell_nnz_         = 0;
+        this->mat_.ELL.max_row = 0;
+        this->coo_nnz_         = 0;
 
-            if(this->coo_nnz_ > 0)
-            {
-                free_hip(&this->mat_.COO.row);
-                free_hip(&this->mat_.COO.col);
-                free_hip(&this->mat_.COO.val);
-
-                this->coo_nnz_ = 0;
-            }
-
-            this->nrow_ = 0;
-            this->ncol_ = 0;
-            this->nnz_  = 0;
-        }
+        this->nrow_ = 0;
+        this->ncol_ = 0;
+        this->nnz_  = 0;
     }
 
     template <typename ValueType>
@@ -212,7 +192,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixHYB<ValueType>*>(&src)) != NULL)
@@ -227,46 +206,16 @@ namespace rocalution
             }
 
             assert(this->nnz_ == cast_mat->nnz_);
+            assert(this->ell_nnz_ == cast_mat->ell_nnz_);
+            assert(this->coo_nnz_ == cast_mat->coo_nnz_);
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->ell_nnz_ > 0)
-            {
-                // ELL
-                hipMemcpy(this->mat_.ELL.col,
-                          cast_mat->mat_.ELL.col,
-                          this->ell_nnz_ * sizeof(int),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.ELL.val,
-                          cast_mat->mat_.ELL.val,
-                          this->ell_nnz_ * sizeof(ValueType),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
-
-            if(this->coo_nnz_ > 0)
-            {
-                // COO
-                hipMemcpy(this->mat_.COO.row,
-                          cast_mat->mat_.COO.row,
-                          (this->coo_nnz_) * sizeof(int),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.COO.col,
-                          cast_mat->mat_.COO.col,
-                          this->coo_nnz_ * sizeof(int),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.COO.val,
-                          cast_mat->mat_.COO.val,
-                          this->coo_nnz_ * sizeof(ValueType),
-                          hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->ell_nnz_, cast_mat->mat_.ELL.col, this->mat_.ELL.col);
+            copy_h2d(this->ell_nnz_, cast_mat->mat_.ELL.val, this->mat_.ELL.val);
+            copy_h2d(this->coo_nnz_, cast_mat->mat_.COO.row, this->mat_.COO.row);
+            copy_h2d(this->coo_nnz_, cast_mat->mat_.COO.col, this->mat_.COO.col);
+            copy_h2d(this->coo_nnz_, cast_mat->mat_.COO.val, this->mat_.COO.val);
         }
         else
         {
@@ -300,46 +249,16 @@ namespace rocalution
             }
 
             assert(this->nnz_ == cast_mat->nnz_);
+            assert(this->ell_nnz_ == cast_mat->ell_nnz_);
+            assert(this->coo_nnz_ == cast_mat->coo_nnz_);
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->ell_nnz_ > 0)
-            {
-                // ELL
-                hipMemcpy(cast_mat->mat_.ELL.col,
-                          this->mat_.ELL.col,
-                          this->ell_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(cast_mat->mat_.ELL.val,
-                          this->mat_.ELL.val,
-                          this->ell_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
-
-            if(this->coo_nnz_ > 0)
-            {
-                // COO
-                hipMemcpy(cast_mat->mat_.COO.row,
-                          this->mat_.COO.row,
-                          this->coo_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(cast_mat->mat_.COO.col,
-                          this->mat_.COO.col,
-                          this->coo_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(cast_mat->mat_.COO.val,
-                          this->mat_.COO.val,
-                          this->coo_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->ell_nnz_, this->mat_.ELL.col, cast_mat->mat_.ELL.col);
+            copy_d2h(this->ell_nnz_, this->mat_.ELL.val, cast_mat->mat_.ELL.val);
+            copy_d2h(this->coo_nnz_, this->mat_.COO.row, cast_mat->mat_.COO.row);
+            copy_d2h(this->coo_nnz_, this->mat_.COO.col, cast_mat->mat_.COO.col);
+            copy_d2h(this->coo_nnz_, this->mat_.COO.val, cast_mat->mat_.COO.val);
         }
         else
         {
@@ -358,7 +277,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixHYB<ValueType>*>(&src)) != NULL)
@@ -373,46 +291,16 @@ namespace rocalution
             }
 
             assert(this->nnz_ == hip_cast_mat->nnz_);
+            assert(this->ell_nnz_ == hip_cast_mat->ell_nnz_);
+            assert(this->coo_nnz_ == hip_cast_mat->coo_nnz_);
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->ell_nnz_ > 0)
-            {
-                // ELL
-                hipMemcpy(this->mat_.ELL.col,
-                          hip_cast_mat->mat_.ELL.col,
-                          this->ell_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.ELL.val,
-                          hip_cast_mat->mat_.ELL.val,
-                          this->ell_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
-
-            if(this->coo_nnz_ > 0)
-            {
-                // COO
-                hipMemcpy(this->mat_.COO.row,
-                          hip_cast_mat->mat_.COO.row,
-                          (this->coo_nnz_) * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.COO.col,
-                          hip_cast_mat->mat_.COO.col,
-                          this->coo_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.COO.val,
-                          hip_cast_mat->mat_.COO.val,
-                          this->coo_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->ell_nnz_, hip_cast_mat->mat_.ELL.col, this->mat_.ELL.col);
+            copy_d2d(this->ell_nnz_, hip_cast_mat->mat_.ELL.val, this->mat_.ELL.val);
+            copy_d2d(this->coo_nnz_, hip_cast_mat->mat_.COO.row, this->mat_.COO.row);
+            copy_d2d(this->coo_nnz_, hip_cast_mat->mat_.COO.col, this->mat_.COO.col);
+            copy_d2d(this->coo_nnz_, hip_cast_mat->mat_.COO.val, this->mat_.COO.val);
         }
         else
         {
@@ -455,46 +343,16 @@ namespace rocalution
             }
 
             assert(this->nnz_ == hip_cast_mat->nnz_);
+            assert(this->ell_nnz_ == hip_cast_mat->ell_nnz_);
+            assert(this->coo_nnz_ == hip_cast_mat->coo_nnz_);
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->ell_nnz_ > 0)
-            {
-                // ELL
-                hipMemcpy(hip_cast_mat->mat_.ELL.col,
-                          this->mat_.ELL.col,
-                          this->ell_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.ELL.val,
-                          this->mat_.ELL.val,
-                          this->ell_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
-
-            if(this->coo_nnz_ > 0)
-            {
-                // COO
-                hipMemcpy(hip_cast_mat->mat_.COO.row,
-                          this->mat_.COO.row,
-                          this->coo_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.COO.col,
-                          this->mat_.COO.col,
-                          this->coo_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.COO.val,
-                          this->mat_.COO.val,
-                          this->coo_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->ell_nnz_, this->mat_.ELL.col, hip_cast_mat->mat_.ELL.col);
+            copy_d2d(this->ell_nnz_, this->mat_.ELL.val, hip_cast_mat->mat_.ELL.val);
+            copy_d2d(this->coo_nnz_, this->mat_.COO.row, hip_cast_mat->mat_.COO.row);
+            copy_d2d(this->coo_nnz_, this->mat_.COO.col, hip_cast_mat->mat_.COO.col);
+            copy_d2d(this->coo_nnz_, this->mat_.COO.val, hip_cast_mat->mat_.COO.val);
         }
         else
         {
@@ -520,7 +378,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // CPU to HIP copy
         if((cast_mat = dynamic_cast<const HostMatrixHYB<ValueType>*>(&src)) != NULL)
@@ -535,46 +392,36 @@ namespace rocalution
             }
 
             assert(this->nnz_ == cast_mat->nnz_);
+            assert(this->ell_nnz_ == cast_mat->ell_nnz_);
+            assert(this->coo_nnz_ == cast_mat->coo_nnz_);
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->ell_nnz_ > 0)
-            {
-                // ELL
-                hipMemcpyAsync(this->mat_.ELL.col,
-                               cast_mat->mat_.ELL.col,
-                               this->ell_nnz_ * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.ELL.val,
-                               cast_mat->mat_.ELL.val,
-                               this->ell_nnz_ * sizeof(ValueType),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
-
-            if(this->coo_nnz_ > 0)
-            {
-                // COO
-                hipMemcpyAsync(this->mat_.COO.row,
-                               cast_mat->mat_.COO.row,
-                               (this->coo_nnz_) * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.COO.col,
-                               cast_mat->mat_.COO.col,
-                               this->coo_nnz_ * sizeof(int),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(this->mat_.COO.val,
-                               cast_mat->mat_.COO.val,
-                               this->coo_nnz_ * sizeof(ValueType),
-                               hipMemcpyHostToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_h2d(this->ell_nnz_,
+                     cast_mat->mat_.ELL.col,
+                     this->mat_.ELL.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_h2d(this->ell_nnz_,
+                     cast_mat->mat_.ELL.val,
+                     this->mat_.ELL.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_h2d(this->coo_nnz_,
+                     cast_mat->mat_.COO.row,
+                     this->mat_.COO.row,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_h2d(this->coo_nnz_,
+                     cast_mat->mat_.COO.col,
+                     this->mat_.COO.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_h2d(this->coo_nnz_,
+                     cast_mat->mat_.COO.val,
+                     this->mat_.COO.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -608,46 +455,36 @@ namespace rocalution
             }
 
             assert(this->nnz_ == cast_mat->nnz_);
+            assert(this->ell_nnz_ == cast_mat->ell_nnz_);
+            assert(this->coo_nnz_ == cast_mat->coo_nnz_);
             assert(this->nrow_ == cast_mat->nrow_);
             assert(this->ncol_ == cast_mat->ncol_);
 
-            if(this->ell_nnz_ > 0)
-            {
-                // ELL
-                hipMemcpyAsync(cast_mat->mat_.ELL.col,
-                               this->mat_.ELL.col,
-                               this->ell_nnz_ * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(cast_mat->mat_.ELL.val,
-                               this->mat_.ELL.val,
-                               this->ell_nnz_ * sizeof(ValueType),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
-
-            if(this->coo_nnz_ > 0)
-            {
-                // COO
-                hipMemcpyAsync(cast_mat->mat_.COO.row,
-                               this->mat_.COO.row,
-                               this->coo_nnz_ * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(cast_mat->mat_.COO.col,
-                               this->mat_.COO.col,
-                               this->coo_nnz_ * sizeof(int),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpyAsync(cast_mat->mat_.COO.val,
-                               this->mat_.COO.val,
-                               this->coo_nnz_ * sizeof(ValueType),
-                               hipMemcpyDeviceToHost);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2h(this->ell_nnz_,
+                     this->mat_.ELL.col,
+                     cast_mat->mat_.ELL.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->ell_nnz_,
+                     this->mat_.ELL.val,
+                     cast_mat->mat_.ELL.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->coo_nnz_,
+                     this->mat_.COO.row,
+                     cast_mat->mat_.COO.row,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->coo_nnz_,
+                     this->mat_.COO.col,
+                     cast_mat->mat_.COO.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2h(this->coo_nnz_,
+                     this->mat_.COO.val,
+                     cast_mat->mat_.COO.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -666,7 +503,6 @@ namespace rocalution
 
         // copy only in the same format
         assert(this->GetMatFormat() == src.GetMatFormat());
-        assert(this->GetMatBlockDimension() == src.GetMatBlockDimension());
 
         // HIP to HIP copy
         if((hip_cast_mat = dynamic_cast<const HIPAcceleratorMatrixHYB<ValueType>*>(&src)) != NULL)
@@ -681,46 +517,36 @@ namespace rocalution
             }
 
             assert(this->nnz_ == hip_cast_mat->nnz_);
+            assert(this->ell_nnz_ == hip_cast_mat->ell_nnz_);
+            assert(this->coo_nnz_ == hip_cast_mat->coo_nnz_);
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->ell_nnz_ > 0)
-            {
-                // ELL
-                hipMemcpy(this->mat_.ELL.col,
-                          hip_cast_mat->mat_.ELL.col,
-                          this->ell_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.ELL.val,
-                          hip_cast_mat->mat_.ELL.val,
-                          this->ell_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
-
-            if(this->coo_nnz_ > 0)
-            {
-                // COO
-                hipMemcpy(this->mat_.COO.row,
-                          hip_cast_mat->mat_.COO.row,
-                          (this->coo_nnz_) * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.COO.col,
-                          hip_cast_mat->mat_.COO.col,
-                          this->coo_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(this->mat_.COO.val,
-                          hip_cast_mat->mat_.COO.val,
-                          this->coo_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->ell_nnz_,
+                     hip_cast_mat->mat_.ELL.col,
+                     this->mat_.ELL.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->ell_nnz_,
+                     hip_cast_mat->mat_.ELL.val,
+                     this->mat_.ELL.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->coo_nnz_,
+                     hip_cast_mat->mat_.COO.row,
+                     this->mat_.COO.row,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->coo_nnz_,
+                     hip_cast_mat->mat_.COO.col,
+                     this->mat_.COO.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->coo_nnz_,
+                     hip_cast_mat->mat_.COO.val,
+                     this->mat_.COO.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -763,46 +589,36 @@ namespace rocalution
             }
 
             assert(this->nnz_ == hip_cast_mat->nnz_);
+            assert(this->ell_nnz_ == hip_cast_mat->ell_nnz_);
+            assert(this->coo_nnz_ == hip_cast_mat->coo_nnz_);
             assert(this->nrow_ == hip_cast_mat->nrow_);
             assert(this->ncol_ == hip_cast_mat->ncol_);
 
-            if(this->ell_nnz_ > 0)
-            {
-                // ELL
-                hipMemcpy(hip_cast_mat->mat_.ELL.col,
-                          this->mat_.ELL.col,
-                          this->ell_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.ELL.val,
-                          this->mat_.ELL.val,
-                          this->ell_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
-
-            if(this->coo_nnz_ > 0)
-            {
-                // COO
-                hipMemcpy(hip_cast_mat->mat_.COO.row,
-                          this->mat_.COO.row,
-                          this->coo_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.COO.col,
-                          this->mat_.COO.col,
-                          this->coo_nnz_ * sizeof(int),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-
-                hipMemcpy(hip_cast_mat->mat_.COO.val,
-                          this->mat_.COO.val,
-                          this->coo_nnz_ * sizeof(ValueType),
-                          hipMemcpyDeviceToDevice);
-                CHECK_HIP_ERROR(__FILE__, __LINE__);
-            }
+            copy_d2d(this->ell_nnz_,
+                     this->mat_.ELL.col,
+                     hip_cast_mat->mat_.ELL.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->ell_nnz_,
+                     this->mat_.ELL.val,
+                     hip_cast_mat->mat_.ELL.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->coo_nnz_,
+                     this->mat_.COO.row,
+                     hip_cast_mat->mat_.COO.row,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->coo_nnz_,
+                     this->mat_.COO.col,
+                     hip_cast_mat->mat_.COO.col,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
+            copy_d2d(this->coo_nnz_,
+                     this->mat_.COO.val,
+                     hip_cast_mat->mat_.COO.val,
+                     true,
+                     HIPSTREAM(_get_backend_descriptor()->HIP_stream_current));
         }
         else
         {
@@ -826,9 +642,11 @@ namespace rocalution
     {
         this->Clear();
 
-        // empty matrix is empty matrix
+        // Empty matrix
         if(mat.GetNnz() == 0)
         {
+            this->AllocateHYB(0, 0, 0, mat.GetM(), mat.GetN());
+
             return true;
         }
 
@@ -846,11 +664,11 @@ namespace rocalution
         {
             this->Clear();
 
-            int nnz_hyb;
-            int nnz_ell;
-            int nnz_coo;
+            int64_t nnz_hyb;
+            int64_t nnz_ell;
+            int64_t nnz_coo;
 
-            if(csr_to_hyb_hip(this->local_backend_.HIP_block_size,
+            if(csr_to_hyb_hip(&this->local_backend_,
                               cast_mat_csr->nnz_,
                               cast_mat_csr->nrow_,
                               cast_mat_csr->ncol_,
