@@ -1,5 +1,5 @@
 /* ************************************************************************
- * Copyright (C) 2018-2022 Advanced Micro Devices, Inc. All rights Reserved.
+ * Copyright (C) 2018-2024 Advanced Micro Devices, Inc. All rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -24,10 +24,14 @@
 #ifndef ROCALUTION_HIP_HIP_SPARSE_HPP_
 #define ROCALUTION_HIP_HIP_SPARSE_HPP_
 
+#include "../../utils/type_traits.hpp"
 #include <rocsparse/rocsparse.h>
 
 namespace rocalution
 {
+    // ValueType to rocsparse_datatype
+    template <typename ValueType>
+    rocsparse_datatype rocsparseTdatatype();
 
     // rocsparse csrmv analysis
     template <typename ValueType>
@@ -104,6 +108,54 @@ namespace rocalution
                                      rocsparse_solve_policy    policy,
                                      void*                     temp_buffer);
 
+    // rocsprarse csritsv buffer size
+    template <typename ValueType>
+    rocsparse_status rocsparseTcsritsv_buffer_size(rocsparse_handle          handle,
+                                                   rocsparse_operation       trans,
+                                                   rocsparse_int             m,
+                                                   rocsparse_int             nnz,
+                                                   const rocsparse_mat_descr descr,
+                                                   const ValueType*          csr_val,
+                                                   const rocsparse_int*      csr_row_ptr,
+                                                   const rocsparse_int*      csr_col_ind,
+                                                   rocsparse_mat_info        info,
+                                                   size_t*                   buffer_size);
+
+    // rocsprarse csritsv analysis
+    template <typename ValueType>
+    rocsparse_status rocsparseTcsritsv_analysis(rocsparse_handle          handle,
+                                                rocsparse_operation       trans,
+                                                rocsparse_int             m,
+                                                rocsparse_int             nnz,
+                                                const rocsparse_mat_descr descr,
+                                                const ValueType*          csr_val,
+                                                const rocsparse_int*      csr_row_ptr,
+                                                const rocsparse_int*      csr_col_ind,
+                                                rocsparse_mat_info        info,
+                                                rocsparse_analysis_policy analysis,
+                                                rocsparse_solve_policy    solve,
+                                                void*                     temp_buffer);
+
+    // rocsprarse csritsv solve
+    template <typename ValueType>
+    rocsparse_status rocsparseTcsritsv_solve(rocsparse_handle                   handle,
+                                             rocsparse_int*                     host_nmaxiter,
+                                             const numeric_traits_t<ValueType>* host_tol,
+                                             numeric_traits_t<ValueType>*       host_history,
+                                             rocsparse_operation                trans,
+                                             rocsparse_int                      m,
+                                             rocsparse_int                      nnz,
+                                             const ValueType*                   alpha,
+                                             const rocsparse_mat_descr          descr,
+                                             const ValueType*                   csr_val,
+                                             const rocsparse_int*               csr_row_ptr,
+                                             const rocsparse_int*               csr_col_ind,
+                                             rocsparse_mat_info                 info,
+                                             const ValueType*                   x,
+                                             ValueType*                         y,
+                                             rocsparse_solve_policy             policy,
+                                             void*                              temp_buffer);
+
     // rocsparse bsrsv buffer size
     template <typename ValueType>
     rocsparse_status rocsparseTbsrsv_buffer_size(rocsparse_handle          handle,
@@ -204,6 +256,28 @@ namespace rocalution
                                      const ValueType*          beta,
                                      ValueType*                y);
 
+    // rocsparse csrgeam
+    template <typename ValueType>
+    rocsparse_status rocsparseTcsrgeam(rocsparse_handle          handle,
+                                       int                       m,
+                                       int                       n,
+                                       const ValueType*          alpha,
+                                       const rocsparse_mat_descr descr_A,
+                                       int                       nnz_A,
+                                       const ValueType*          csr_val_A,
+                                       const int*                csr_row_ptr_A,
+                                       const int*                csr_col_ind_A,
+                                       const ValueType*          beta,
+                                       const rocsparse_mat_descr descr_B,
+                                       int                       nnz_B,
+                                       const ValueType*          csr_val_B,
+                                       const int*                csr_row_ptr_B,
+                                       const int*                csr_col_ind_B,
+                                       const rocsparse_mat_descr descr_C,
+                                       ValueType*                csr_val_C,
+                                       const int*                csr_row_ptr_C,
+                                       int*                      csr_col_ind_C);
+
     // rocsparse csrgemm
     template <typename ValueType>
     rocsparse_status rocsparseTcsrgemm_buffer_size(rocsparse_handle          handle,
@@ -564,6 +638,32 @@ namespace rocalution
                                                 rocsparse_int*            csr_col_ind_C,
                                                 ValueType                 tol);
 
+    // rocsparse csritilu0 compute
+    template <typename ValueType>
+    rocsparse_status rocsparseTcsritilu0_compute(rocsparse_handle            handle,
+                                                 rocsparse_itilu0_alg        alg,
+                                                 rocsparse_int               option,
+                                                 rocsparse_int*              nmaxiter,
+                                                 numeric_traits_t<ValueType> tol,
+                                                 rocsparse_int               m,
+                                                 rocsparse_int               nnz,
+                                                 const rocsparse_int*        csr_row_ptr,
+                                                 const rocsparse_int*        csr_col_ind,
+                                                 const ValueType*            csr_val,
+                                                 ValueType*                  ilu0,
+                                                 rocsparse_index_base        idx_base,
+                                                 size_t                      buffer_size,
+                                                 void*                       buffer);
+
+    // rocsparse csritilu0 history
+    template <typename ValueType>
+    rocsparse_status rocsparseTcsritilu0_history(rocsparse_handle             handle,
+                                                 rocsparse_itilu0_alg         alg,
+                                                 rocsparse_int*               niter,
+                                                 numeric_traits_t<ValueType>* data,
+                                                 size_t                       buffer_size,
+                                                 void*                        buffer);
+
 } // namespace rocalution
 
 #endif // ROCALUTION_HIP_HIP_SPARSE_HPP_
