@@ -1,5 +1,5 @@
 /* ************************************************************************
- * Copyright (C) 2018-2022 Advanced Micro Devices, Inc. All rights Reserved.
+ * Copyright (C) 2018-2023 Advanced Micro Devices, Inc. All rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -25,6 +25,7 @@
 #define ROCALUTION_HIP_HIP_UTILS_HPP_
 
 #include "../../utils/log.hpp"
+#include "../../utils/type_traits.hpp"
 #include "../backend_manager.hpp"
 #include "backend_hip.hpp"
 #include "hip_atomics.hpp"
@@ -40,8 +41,7 @@
 
 // clang-format off
 #ifndef ROCALUTION_USE_MOVE_DPP
-#if defined(__gfx803__) || defined(__gfx900__) || defined(__gfx906__) || defined(__gfx908__) \
-    || defined(__gfx90a__)
+#if defined(__GFX8__) || defined(__GFX9__)
 #define ROCALUTION_USE_MOVE_DPP 1
 #else
 #define ROCALUTION_USE_MOVE_DPP 0
@@ -51,6 +51,7 @@
 
 #define ROCBLAS_HANDLE(handle) *static_cast<rocblas_handle*>(handle)
 #define ROCSPARSE_HANDLE(handle) *static_cast<rocsparse_handle*>(handle)
+#define HIPSTREAM(handle) *static_cast<hipStream_t*>(handle)
 
 #define CHECK_HIP_ERROR(file, line)                              \
     {                                                            \
@@ -216,8 +217,6 @@ namespace rocalution
     }
 #endif
 
-    __device__ int __llvm_amdgcn_readlane(int index, int offset) __asm("llvm.amdgcn.readlane");
-
 #if ROCALUTION_USE_MOVE_DPP
     template <unsigned int WFSIZE>
     static __device__ __forceinline__ void wf_reduce_sum(int* sum)
@@ -291,7 +290,7 @@ namespace rocalution
 
         if(WF_SIZE > 32)
         {
-            upper_sum.b32 = __llvm_amdgcn_readlane(temp_sum.b32, 32);
+            upper_sum.b32 = __builtin_amdgcn_readlane(temp_sum.b32, 32);
             temp_sum.val += upper_sum.val;
         }
 
@@ -348,8 +347,8 @@ namespace rocalution
 
         if(WF_SIZE > 32)
         {
-            upper_sum.b32[0] = __llvm_amdgcn_readlane(temp_sum.b32[0], 32);
-            upper_sum.b32[1] = __llvm_amdgcn_readlane(temp_sum.b32[1], 32);
+            upper_sum.b32[0] = __builtin_amdgcn_readlane(temp_sum.b32[0], 32);
+            upper_sum.b32[1] = __builtin_amdgcn_readlane(temp_sum.b32[1], 32);
             temp_sum.val += upper_sum.val;
         }
 
@@ -408,7 +407,6 @@ namespace rocalution
     static __device__ __forceinline__ void wf_reduce_min(std::complex<ValueType>* val)
     {
         ValueType real = std::real(*val);
-        ValueType imag = std::imag(*val);
 
         wf_reduce_min<WF_SIZE>(&real);
 
@@ -430,7 +428,6 @@ namespace rocalution
     static __device__ __forceinline__ void wf_reduce_max(std::complex<ValueType>* val)
     {
         ValueType real = std::real(*val);
-        ValueType imag = std::imag(*val);
 
         wf_reduce_max<WF_SIZE>(&real);
 
@@ -651,18 +648,18 @@ namespace rocalution
 
     template <unsigned int BLOCKSIZE, typename IndexType>
     __launch_bounds__(BLOCKSIZE) __global__
-        void kernel_find_maximum_blockreduce(IndexType m,
+        void kernel_find_maximum_blockreduce(int64_t m,
                                              const IndexType* __restrict__ array,
                                              IndexType* __restrict__ workspace)
     {
-        IndexType tid = hipThreadIdx_x;
-        IndexType gid = hipBlockIdx_x * BLOCKSIZE + tid;
+        unsigned int tid = hipThreadIdx_x;
+        int64_t      gid = hipBlockIdx_x * BLOCKSIZE + tid;
 
         __shared__ IndexType smax[BLOCKSIZE];
 
         IndexType t_max = -2;
 
-        for(IndexType idx = gid; idx < m; idx += hipGridDim_x * BLOCKSIZE)
+        for(int64_t idx = gid; idx < m; idx += hipGridDim_x * BLOCKSIZE)
         {
             t_max = max(t_max, array[idx]);
         }
@@ -681,9 +678,9 @@ namespace rocalution
 
     template <unsigned int BLOCKSIZE, typename IndexType>
     __launch_bounds__(BLOCKSIZE) __global__
-        void kernel_find_maximum_finalreduce(IndexType m, IndexType* __restrict__ workspace)
+        void kernel_find_maximum_finalreduce(IndexType* __restrict__ workspace)
     {
-        IndexType tid = hipThreadIdx_x;
+        unsigned int tid = hipThreadIdx_x;
 
         __shared__ IndexType sdata[BLOCKSIZE];
 
@@ -699,6 +696,21 @@ namespace rocalution
         }
     }
 
+    template <unsigned int BLOCKSIZE, typename ValueType>
+    __launch_bounds__(BLOCKSIZE) __global__ void kernel_axpy(int64_t   size,
+                                                             ValueType alpha,
+                                                             const ValueType* __restrict__ x,
+                                                             ValueType* __restrict__ y)
+    {
+        unsigned int tid = hipThreadIdx_x;
+        int64_t      gid = hipBlockIdx_x * BLOCKSIZE + tid;
+
+        if(gid < size)
+        {
+            y[gid] += alpha * x[gid];
+        }
+    }
+
     static __device__ __forceinline__ float hip_shfl_xor(float val, int i)
     {
         return __shfl_xor(val, i);
@@ -720,114 +732,6 @@ namespace rocalution
     }
 #endif
 
-    // Hash related functionality
-    // Insert key, returns true if key has been added, false if key is already present
-    template <unsigned int HASHVAL, unsigned int HASHSIZE, typename IndexType>
-    static __device__ __forceinline__ bool
-        insert_key(IndexType key, IndexType* table, IndexType empty)
-    {
-        if(key == empty)
-        {
-            return false;
-        }
-
-        // Compute hash
-        IndexType hash = (key * HASHVAL) & (HASHSIZE - 1);
-
-        // Loop until key has been inserted
-        while(true)
-        {
-            if(table[hash] == key)
-            {
-                // Element already present
-                return false;
-            }
-            else if(table[hash] == empty)
-            {
-                // If empty, add element with atomic
-                if(atomicCAS(&table[hash], empty, key) == empty)
-                {
-                    // Increment number of insertions
-                    return true;
-                }
-            }
-            else
-            {
-                // Linear probing, when hash is collided, try next entry
-                hash = (hash + 1) & (HASHSIZE - 1);
-            }
-        }
-    }
-
-    // Hash related functionality
-    // Appends value to existing key, returns true if appended, false else
-    template <unsigned int HASHVAL, unsigned int HASHSIZE, typename IndexType, typename ValueType>
-    static __device__ __forceinline__ bool append_pair(
-        IndexType key, ValueType val, IndexType* table, ValueType* data, IndexType empty)
-    {
-        if(key == empty)
-        {
-            return false;
-        }
-
-        // Compute hash
-        IndexType hash = (key * HASHVAL) & (HASHSIZE - 1);
-
-        // Loop until key has been inserted
-        while(true)
-        {
-            if(table[hash] == key)
-            {
-                // Element already present
-                atomicAdd(&data[hash], val);
-                return true;
-            }
-            else if(table[hash] == empty)
-            {
-                // If empty, nothing to append
-                return false;
-            }
-            else
-            {
-                // Linear probing, when hash is collided, try next entry
-                hash = (hash + 1) & (HASHSIZE - 1);
-            }
-        }
-    }
-
-    // Hash related functionality
-    // Checks for key, returns true if key exists, false else
-    template <unsigned int HASHVAL, unsigned int HASHSIZE, typename IndexType>
-    static __device__ __forceinline__ bool
-        key_exists(IndexType key, IndexType* table, IndexType empty)
-    {
-        if(key == empty)
-        {
-            return false;
-        }
-
-        // Compute hash
-        IndexType hash = (key * HASHVAL) & (HASHSIZE - 1);
-
-        // Loop until key has been inserted
-        while(true)
-        {
-            if(table[hash] == key)
-            {
-                // Element already present
-                return true;
-            }
-            else if(table[hash] == empty)
-            {
-                return false;
-            }
-            else
-            {
-                // Linear probing, when hash is collided, try next entry
-                hash = (hash + 1) & (HASHSIZE - 1);
-            }
-        }
-    }
 } // namespace rocalution
 
 #endif // ROCALUTION_HIP_HIP_UTILS_HPP_
